##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
InstrumentControl Auxiliary
***************************

:module: instrument_control_auxiliary

:synopsis: Auxiliary used to communicate via a VISA connector using the
    SCPI protocol.

.. currentmodule:: instrument_control_auxiliary

"""
import logging
import queue
import re
import time
from typing import Any, List, Optional, Tuple, Union

from pykiso import CChannel
from pykiso.auxiliary import AuxiliaryInterface, close_connector

from .lib_scpi_commands import LibSCPI

log = logging.getLogger(__name__)


class InstrumentControlAuxiliary(AuxiliaryInterface):
    """Auxiliary used to communicate via a VISA connector using the SCPI
    protocol.
    """

    def __init__(
        self,
        com: CChannel,
        instrument="",
        write_termination="\n",
        output_channel: int = None,
        **kwargs,
    ):
        """Constructor.

        :param com: VISAChannel that supports VISA communication
        :param instrument: name of the instrument currently in use
            (will be used to adapt the SCPI commands)
        :param write_termination: write termination character
        :param output_channel: output channel to use on the instrument
            currently in use (if more than one)
        """
        super().__init__(
            is_proxy_capable=False,
            tx_task_on=False,
            rx_task_on=False,
            **kwargs,
        )
        self.channel = com
        self.instrument = instrument
        self.write_termination = write_termination
        self.output_channel = output_channel
        self.helpers = LibSCPI(self, self.instrument)

    def write(
        self,
        write_command: str,
        validation: Tuple[str, Union[str, List[str]]] = None,
    ) -> str:
        """Send a write request to the instrument.

        :param write_command: command to send
        :param validation: contain validation criteria apply on the
            response
        """
        log.internal_debug(f"Sending a write request in {self} for {write_command}")
        return self.handle_write(write_command, validation)

    def handle_write(
        self,
        write_command: str,
        validation: Tuple[str, Union[str, List[str]]] = None,
    ) -> str:
        """Send a write request to the instrument and then returns if
        the value was successfully written. A query is sent immediately
        after the writing and the answer is compared to the expected
        one.

        :param write_command: write command to send
        :param validation: tuple of the form
            (validation command (str), expected output (str or list of str))

        :return: status message depending on the command validation:
            SUCCESS, FAILURE or NO_VALIDATION
        """
        log.internal_debug(f"Sending a write request in {self} for {write_command}")
        # Send the message with the termination character
        self.channel.cc_send(msg=write_command + self.write_termination)

        if validation is not None:
            # Check that the writing request was successfully performed on the instrument
            validation_query, validation_expected_output = validation
            if isinstance(validation_expected_output, str):
                validation_expected_output = [validation_expected_output]

            # Send the validation query with a delay between the previous write and this query request
            # to make sure that the instrument has had sufficient time to complete the operation
            time.sleep(0.1)
            validation_query_response = self.handle_query(validation_query)

            # Check that the responses matches the expected response
            if isinstance(validation_query_response, str) and validation_query_response != "":
                # Check the tag VALUE{} in the expected output and adapt the validation process
                match = re.compile(r"VALUE{([+-]?\d+[\.\d]*)}").findall(validation_expected_output[0])
                if match:
                    # tag "VALUE{value}" detected. This tag can be used when we want to ensure that
                    # a numerical parameter was correctly set on the instrument
                    if float(match[0]) == float(validation_query_response.split()[0]):
                        write_success = True
                    else:
                        write_success = False
                else:
                    # compare the expected output with the result of the validation query command
                    if validation_query_response in validation_expected_output:
                        write_success = True
                    else:
                        write_success = False
                if write_success is True:
                    log.internal_debug(f"Write request {write_command} successful after verification")
                    return "SUCCESS"
                else:
                    log.internal_warning(
                        f"Write request {write_command} failed! Validation query response was different than expected."
                    )
                    return "FAILURE"
            else:
                log.internal_warning(
                    f"Write request {write_command} failed! No response received for the validation query."
                )
                return "FAILURE"
        else:
            log.internal_debug(f"Write request {write_command} processed without validation")
            return "NO_VALIDATION"

    def read(self) -> Union[str, bool]:
        """Send a read request to the instrument.

        :return: received response from instrument otherwise empty
            string
        """
        log.internal_debug(f"Sending a read request in {self}")
        return self.handle_read()

    def handle_read(self) -> Optional[str]:
        """Handle read command by calling associated connector
        cc_receive.

        :return: received response from instrument otherwise empty
            string
        """
        response = self.channel.cc_receive()
        response_data = response.get("msg")
        if isinstance(response_data, bytes):
            response_data = response_data.decode().strip()
        return response_data

    def query(self, query_command: str) -> Union[bytes, str]:
        """Send a query request to the instrument. Uses the 'query' method of the
            channel if available, uses 'cc_send' and 'cc_receive' otherwise.

        :param query_command: query command to send

        :return: Response message, None if the request expired with a
            timeout.
        """
        log.internal_debug(f"Sending a query request in {self}) for {query_command}")
        return self.handle_query(query_command)

    def handle_query(self, query_command: str) -> Optional[str]:
        """Send a query request to the instrument. Uses the 'query' method of the
            channel if available, uses 'cc_send' and 'cc_receive' otherwise.

        :param query_command: query command to send

        :return: Response message, None if the request expired with a
            timeout.
        """
        if hasattr(self.channel, "query"):
            response = self.channel.query(query_command + self.write_termination)
            return response.get("msg")
        else:
            self.channel.cc_send(msg=query_command + self.write_termination)
            time.sleep(0.05)
            response = self.channel.cc_receive()
            response_data = response.get("msg")
            if isinstance(response_data, bytes):
                response_data = response_data.decode().strip()
            return response_data

    def _create_auxiliary_instance(self) -> bool:
        """Open the connector.

        :return: always True
        """
        log.internal_info("Create auxiliary instance")

        try:
            # Open instrument
            log.internal_info("Open instrument")
            self.channel.open()

            # Enable remote control
            log.internal_info("Enable remote control")
            command, validation = self.helpers.get_command(
                cmd_tag="REMOTE_CONTROL",
                cmd_type="write",
                cmd_validation=["REMOTE", "ON", "1"],
            )
            self.handle_write(f"{command} ON".strip(), validation)

            # Select channel if needed
            if self.output_channel is not None:
                log.internal_info(f"Select channel {self.output_channel}")
                command, validation = self.helpers.get_command(
                    cmd_tag="OUTPUT_CHANNEL",
                    cmd_type="write",
                    cmd_validation=f"{self.output_channel}",
                )
                self.handle_write(f"{command} {self.output_channel}".strip(), validation)
            else:
                log.internal_info("Using default output channel.")
        except Exception:
            log.exception("Unable to safely open the instrument.")
            return False
        return True

    @close_connector
    def _delete_auxiliary_instance(self) -> bool:
        """Close the connector communication.

        :return: True if the connectors is closed otherwise False
        """
        log.internal_info("Auxiliary instance deleted")
        return True

    def _run_command(self, cmd_message: Any, cmd_data: Optional[bytes]) -> None:
        """Not used.

        Simply respect the interface.
        """

    def _receive_message(self, timeout_in_s: float) -> None:
        """Not used.

        Simply respect the interface.
        """



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
uds_base_auxiliary
******************

:module: uds_base_auxiliary

:synopsis: Auxiliary base class providing common methods for
    Unified Diagnostic Service protocol handling.
    This base auxiliary is not meant to be used directly, use its
    subclasses instead.

.. currentmodule:: uds_base_auxiliary

"""
import logging
import warnings
from pathlib import Path
from typing import Optional, Union

from uds import Config, Uds

from pykiso.auxiliary import AuxiliaryInterface, close_connector, open_connector
from pykiso.connector import CChannel

log = logging.getLogger(__name__)


class UdsBaseAuxiliary(AuxiliaryInterface):
    """Base Auxiliary class for handling UDS protocol."""

    POSITIVE_RESPONSE_OFFSET = 0x40
    DEFAULT_UDS_CONFIG = {
        "transport_protocol": "CAN",
        "p2_can_client": 5,
        "p2_can_server": 1,
    }
    DEFAULT_TP_CONFIG = {
        "addressing_type": "NORMAL",
        "n_sa": 0xFF,
        "n_ta": 0xFF,
        "n_ae": 0xFF,
        "m_type": "DIAGNOSTICS",
        "discard_neg_resp": False,
    }

    def __init__(
        self,
        com: CChannel,
        config_ini_path: Optional[Union[Path, str]] = None,
        odx_file_path: Optional[Union[Path, str]] = None,
        request_id: Optional[int] = None,
        response_id: Optional[int] = None,
        tp_layer: dict = None,
        uds_layer: dict = None,
        **kwargs,
    ):
        """Initialize attributes.

        :param com: communication channel connector.
        :param config_ini_path: unused. Legacy UDS parameter file.
        :param odx_file_path: ecu diagnostic definition file.
        :param request_id: optional CAN ID used for sending messages.
        :param response_id: optional CAN ID used for receiving messages.
        """
        super().__init__(is_proxy_capable=True, tx_task_on=False, rx_task_on=True, **kwargs)
        self.channel = com
        self.odx_file_path = odx_file_path
        self.req_id = request_id
        self.res_id = response_id
        self.uds_config_enable = False
        self.uds_config = None
        self.tp_layer = None
        self.tp_waiting_time = 0.010
        self.uds_layer = None
        self._init_com_layers(tp_layer, uds_layer)

        # remove it after a few releases just warn users
        if config_ini_path is not None:
            warnings.warn(
                "The usage of ini file is deprecated, nothing will be read from it",
                category=FutureWarning,
            )
            log.warning(f"apply following values for ISOTP layer configuration : {UdsBaseAuxiliary.DEFAULT_TP_CONFIG}")
            log.warning(f"apply following values for UDS layer configuration : {UdsBaseAuxiliary.DEFAULT_UDS_CONFIG}")

    def _init_com_layers(self, tp_layer: Optional[dict] = None, uds_layer: Optional[dict] = None) -> None:
        """Initialize isotp and uds layer attributes.

        :param tp_layer: isotp configuration given at yaml level
        :param uds_layer: uds configuration given at yaml level
        """
        tp_layer = tp_layer or UdsBaseAuxiliary.DEFAULT_TP_CONFIG.copy()
        uds_layer = uds_layer or UdsBaseAuxiliary.DEFAULT_UDS_CONFIG.copy()
        # add configured request id and response id to tp layer config
        tp_layer["req_id"] = self.req_id
        tp_layer["res_id"] = self.res_id
        self.tp_layer = tp_layer
        self.uds_layer = uds_layer

    @open_connector
    def _create_auxiliary_instance(self) -> bool:
        """Open current associated channel.

        :return: if channel creation is successful return True
            otherwise false
        """

        try:
            is_not_none = self.odx_file_path is not None
            is_not_bool = not isinstance(self.odx_file_path, bool)

            if is_not_none and is_not_bool and Path(self.odx_file_path).exists():
                self.uds_config_enable = True
            else:
                self.odx_file_path = None

            Config.load_com_layer_config(self.tp_layer, self.uds_layer)
            self.uds_config = Uds(
                self.odx_file_path,
                connector=self.channel,
            )
            if hasattr(self, "transmit"):
                self.uds_config.overwrite_transmit_method(self.transmit)
            if hasattr(self, "receive"):
                self.uds_config.overwrite_receive_method(self.receive)
            return True
        except Exception:
            log.exception("An error occurred during kiso-python-uds initialization")
            return False

    @close_connector
    def _delete_auxiliary_instance(self) -> bool:
        """Close current associated channel.

        :return: always True
        """
        log.internal_info("Auxiliary instance deleted")
        return True



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Proxy Auxiliary
***************

:module: proxy_auxiliary

:synopsis: auxiliary use to connect multiple auxiliaries on a unique connector.

This auxiliary simply spread all commands and received messages to all connected
auxiliaries. This auxiliary is only usable through proxy connector.

.. code-block:: none

     ___________   ___________         ___________
    |           | |           | ..... |           |
    |   Aux 1   | |   Aux 1   |       |   Aux n   |
    |___________| |___________|       |___________|
          |             |                   |
          |             |                   |
     ___________   ___________         ___________
    |           | |           | ..... |           |
    |Proxy_con 1| |Proxy_con 2|       |Proxy_con n|
    |___________| |___________|       |___________|
          |             |                   |
          |             |                   |
          |             |                   |
     _____________________________________________
    |                                             |
    |               Proxy Auxiliary               |
    |_____________________________________________|
                        |
                        |
     _____________________________________________
    |                                             |
    |               Connector Channel             |
    |_____________________________________________|

.. currentmodule:: proxy_auxiliary

"""

import logging
import queue
import sys
import time
from pathlib import Path
from typing import List, Optional, Tuple

from pykiso import CChannel
from pykiso.auxiliary import AuxiliaryInterface, close_connector, open_connector
from pykiso.lib.connectors.cc_proxy import CCProxy
from pykiso.test_setup.config_registry import ConfigRegistry
from pykiso.test_setup.dynamic_loader import PACKAGE

log = logging.getLogger(__name__)


class ProxyAuxiliary(AuxiliaryInterface):
    """Proxy auxiliary for multi auxiliaries communication handling."""

    def __init__(
        self,
        com: CChannel,
        aux_list: List[str],
        activate_trace: bool = False,
        trace_dir: Optional[str] = None,
        trace_name: Optional[str] = None,
        **kwargs,
    ):
        """Initialize attributes.

        :param com: Communication connector
        :param aux_list: list of auxiliary's alias
        :param activate_trace: log all received messages in a
            dedicated trace file or not
        :param trace_dir: where to place the trace
        :param trace_name: trace's file name
        """
        super().__init__(is_proxy_capable=True, tx_task_on=False, rx_task_on=True, **kwargs)
        self.channel = com
        self._open_count = 0
        self.logger = self._init_trace(activate_trace, trace_dir, trace_name)
        self.proxy_channels = self.get_proxy_con(aux_list)

    def _count_open_proxy_channels(self) -> int:
        """
        Get the number of proxy channels connected to this auxiliary
        that are currently open.
        """
        return len([ccproxy for ccproxy in self.proxy_channels if isinstance(ccproxy.queue_out, queue.Queue)])

    @property
    def _open_connections(self) -> int:
        """A counter monitoring the number of attached running auxiliaries.

        .. warning::
            Do not set this manually as it can easily result in unexpected behaviours.


        :getter: returns the number of currently attached and running auxiliaries.
        :setter: set by ``CCProxy`` instances when being opened or closed.
            When no open connection remains and the counter falls back to 0,
            the ``ProxyAuxiliary`` will be stopped and the attached 'physical'
            communication channel closed.
            When the first connection is opened and the counter is increased
            from 0 to 1, the ``ProxyAuxiliary`` will be started and the attached
            'physical' communication channel opened.
        """
        with self.lock:
            return self._open_count

    @_open_connections.setter
    def _open_connections(self, value: int):
        # locking shouldn't be necessary but we're never too paranoid
        with self.lock:
            # on the original proxy setup, the auxiliaries are created before the proxy
            # therefore, when the first auxiliary will be stopped, the counter value will be -1
            # and if an auxiliary is started then stopped the counter will be 0
            if value <= 0:
                value = self._count_open_proxy_channels()
            last_connection_closed = value == 0 and self._open_count == 1
            first_connection_opened = value == 1 and self._open_count == 0
            self._open_count = value
        # stop proxy if the last attached auxiliary was stopped
        if last_connection_closed and self.is_instance:
            self.delete_instance()
        # start proxy if the first attached auxiliary is started
        elif first_connection_opened and not self.is_instance:
            self.create_instance()

    @staticmethod
    def _init_trace(
        activate: bool,
        t_dir: Optional[str] = None,
        t_name: Optional[str] = None,
    ) -> logging.Logger:
        """Initialize the logging trace for proxy auxiliary received
        message recording.

        :param activate: True if the trace is activate otherwise False
        :param t_dir: trace directory path (absolute or relative)
        :param t_name: trace full name (without file extension)

        :return : created logger containing the configured
            FileHander otherwise default logger
        """
        logger = log

        if not activate:
            return logger

        # Just avoid the case the given trace directory is None
        t_dir = "" if t_dir is None else t_dir
        # if the given log path is not absolute add root path
        # (where pykiso is launched) otherwise take it as it is
        dir_path = (Path() / t_dir).resolve() if not Path(t_dir).is_absolute() else Path(t_dir)
        # if no specific logging file name is given take the default one
        t_name = (
            time.strftime(f"%Y-%m-%d_%H-%M-%S_{t_name}.log")
            if t_name is not None
            else time.strftime("%Y-%m-%d_%H-%M-%S_proxy_logging.log")
        )
        # if path doesn't exists take root path (where pykiso is launched)
        log_path = dir_path / t_name if dir_path.exists() else (Path() / t_name).resolve()

        # configure the file handler and create the trace file
        log_format = logging.Formatter("%(asctime)s : %(message)s")
        log.internal_info(f"create proxy trace file at {log_path}")
        handler = logging.FileHandler(log_path, "w+")
        handler.setFormatter(log_format)
        # create logger and set the log level to DEBUG
        logger = logging.getLogger(f"{__name__}.PROXY")
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

        return logger

    def get_proxy_con(self, aux_list: List[str]) -> Tuple[CCProxy, ...]:
        """Retrieve all connector associated to all given existing Auxiliaries.

        If auxiliary alias exists but auxiliary instance was not created
        yet, create it immediately using ConfigRegistry _aux_cache.

        :param aux_list: list of auxiliary aliases or instances.
        :return: tuple containing all connectors associated to
            all given auxiliaries.
        """
        channel_inst: List[CCProxy] = []

        for aux in aux_list:
            # aux_list can contain a auxiliary instance just grab the
            # channel
            if isinstance(aux, AuxiliaryInterface):
                self._check_aux_compatibility(aux)
                channel_inst.append(aux.channel)
                continue
            # check the system module in order to get the auxiliary
            # instance
            aux_inst = sys.modules.get(f"{PACKAGE}.auxiliaries.{aux}")
            if aux_inst is not None:
                self._check_aux_compatibility(aux_inst)
                channel_inst.append(aux_inst.channel)
            # check if the given aux_name is in the available aux
            # alias list
            elif aux in ConfigRegistry.get_auxes_alias():
                log.internal_info(f"Auxiliary '{aux}' is not using import magic mechanism (pre-loaded)")
                # load it using ConfigRegistry _aux_cache
                aux_inst = ConfigRegistry.get_aux_by_alias(aux)
                self._check_aux_compatibility(aux_inst)
                channel_inst.append(aux_inst.channel)
            # the given auxiliary alias doesn't exist or refer to a
            # invalid one
            else:
                log.error(f"Auxiliary '{aux}' doesn't exist")

        # Check if auxes/connectors are compatible with the proxy aux
        self._check_channels_compatibility(channel_inst)

        # Finally bind the physical channel to the proxy channels to
        # share its API to the user's auxiliaries
        for channel in channel_inst:
            channel._bind_channel_info(self)

        return tuple(channel_inst)

    @staticmethod
    def _check_aux_compatibility(aux: AuxiliaryInterface) -> None:
        """Check if the given auxiliary is proxy compatible.

        :param aux: auxiliary instance to check

        :raises NotImplementedError: if is_proxy_capable flag is False
        """
        if not aux.is_proxy_capable:
            raise NotImplementedError(f"Auxiliary {aux} is not compatible with a proxy auxiliary!")

    @staticmethod
    def _check_channels_compatibility(channels: List[CChannel]) -> None:
        """Check if all associated channels are compatible.

        :param channels: all channels collected by the proxy aux

        :raises TypeError: if the connector is not an instance of
            CCProxy
        """
        for channel in channels:
            if not isinstance(channel, CCProxy):
                raise TypeError(f"Channel {channel} is not compatible!")

    def _dispatch_tx_method_to_channels(self) -> None:
        """Attached public run_command method to all connected proxy
        channels.

        .. note:: This method use the thread safe method implemented by
            each proxy channel (attach_tx_callback).
        """
        for conn in self.proxy_channels:
            conn.attach_tx_callback(self.run_command)

    def _remove_tx_method_from_channels(self) -> None:
        """Detach public run_command method from all connected proxy
        channels.

        .. note:: This method use the thread safe method implemented by
            each proxy channel (detach_tx_callback).
        """
        for conn in self.proxy_channels:
            conn.detach_tx_callback()

    @open_connector
    def _create_auxiliary_instance(self) -> bool:
        """Open current associated channel and dispatch tx method.

        :return: if channel creation is successful return True otherwise
            False
        """
        self._dispatch_tx_method_to_channels()
        log.internal_info("Auxiliary instance created")
        return True

    @close_connector
    def _delete_auxiliary_instance(self) -> bool:
        """Close current associated channel.

        :return: if channel deletion is successful return True otherwise
            False
        """
        self._remove_tx_method_from_channels()
        log.internal_info("Auxiliary instance deleted")
        return True

    def run_command(self, conn: CChannel, *args: tuple, **kwargs: dict) -> None:
        """Transmit an incoming request from a linked proxy channel
        to the proxy auxiliary's channel.

        :param conn: current proxy channel instance which the command
            comes from
        :param args: postional arguments
        :param args: named arguments
        """
        with self.lock:
            self._run_command(conn, *args, **kwargs)

    def _run_command(self, conn: CChannel, *args: tuple, **kwargs: dict) -> None:
        """Send the request coming from the given proxy channel and
        dispatch it to the other linked proxy channels.

        :param conn: current proxy channel instance which the command
            comes from
        :param args: postional arguments
        :param args: named arguments

        In addition, all commands are dispatch to others auxiliaries
        using proxy connector queue out.
        """
        self.channel.cc_send(*args, **kwargs)
        self._dispatch_command(con_use=conn, **kwargs)

    def _dispatch_command(self, con_use: CChannel, **kwargs: dict):
        """Dispatch the current command to others connected auxiliaries.

        This action is performed by populating the queue out from each
        proxy connectors.

        :param con_use: current proxy channel instance which the command
            comes from
        :param kwargs: named arguments
        """
        for conn in self.proxy_channels:
            if conn != con_use and conn.queue_out is not None:
                conn.queue_out.put(kwargs)

    def _receive_message(self, timeout_in_s: float = 0) -> None:
        """Get a message from the associated channel and dispatch it to
        the linked proxy channels.

        .. note:: this method is called by the rx thread task from the
            inherited interface class


        :param timeout_in_s: maximum amount of time (seconds) to wait
            for a message
        """
        try:
            recv_response = self.channel.cc_receive(timeout=timeout_in_s)
            received_data = recv_response.get("msg")
            # if data are received, populate connector's queue_out
            if received_data is not None:
                self.logger.debug(
                    "received response : data %s || channel : %s",
                    received_data,
                    self.channel.name,
                )
                for conn in self.proxy_channels:
                    if conn.queue_out is not None:
                        conn.queue_out.put(recv_response)
        except Exception:
            log.exception(f"encountered error while receiving message via {self.channel}")



##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
UDS Auxiliary acting as a Server/ECU
************************************

:module: uds_server_auxiliary

:synopsis: Auxiliary used to handle Unified Diagnostic Service protocol as a Server.
    This auxiliary is meant to run in the background and replies to configured requests.

.. currentmodule:: uds_auxiliary

"""
from __future__ import annotations

import logging
import threading
import time
from typing import Callable, Dict, List, Optional, Union

from uds import IsoServices

from pykiso.types import OdxRequestConfigDict

from .common.odx_parser import OdxParser
from .common.uds_base_auxiliary import UdsBaseAuxiliary
from .common.uds_callback import UdsCallback
from .common.uds_response import UdsResponse

log = logging.getLogger(__name__)

# possible data lengths for CAN FD padding
CAN_FD_DATA_LENGTHS = (8, 12, 16, 20, 24, 32, 48, 64)


class UdsServerAuxiliary(UdsBaseAuxiliary):
    """Auxiliary used to handle the UDS protocol on server (ECU) side."""

    CAN_FD_PADDING_PATTERN = 0xCC
    services = IsoServices

    def __init__(self, *args, **kwargs):
        """Initialize attributes.

        :param com: communication channel connector.
        :param config_ini_path: uds parameters file.
        :param request_id: optional CAN ID used for sending messages.
        :param response_id: optional CAN ID used for receiving messages.
        :param odx_file_path: ecu diagnostic definition file.
        """
        super().__init__(*args, **kwargs)

        self._ecu_config = None
        if self.odx_file_path is not None:
            self.odx_parser = OdxParser(self.odx_file_path)

        self._callbacks: Dict[str, UdsCallback] = {}
        self._callback_lock = threading.Lock()

    @property
    def callbacks(self):
        """Access the callback dictionary in a thread-safe way.

        :return: the internal callbacks dictionary.
        """
        with self._callback_lock:
            return self._callbacks

    def _create_auxiliary_instance(self) -> bool:
        """Open communication channel, create UDS instance and adapt
        the underlying CAN TP's padding pattern.

        :return: True if creation succeeded otherwise False.
        """
        creation_status = super()._create_auxiliary_instance()
        if not creation_status:
            return False
        self.uds_config.tp.PADDING_PATTERN = self.CAN_FD_PADDING_PATTERN
        return True

    @staticmethod
    def format_data(uds_data: List[int]) -> str:
        """Format UDS data as a list of integers to a hexadecimal string.

        :param uds_data: UDS data as a list of integers.

        :return: the UDS data as a hexadecimal string.
        """
        return f"0x{bytes(uds_data).hex().upper()}"

    @classmethod
    def _pad_message(cls, message: List[int]) -> List[int]:
        """Pad a CAN FD message to send with the configured padding pattern.

        :param message: message to pad.

        :return: the padded message.
        """
        padded_length = next(size for size in CAN_FD_DATA_LENGTHS if size >= len(message))
        return message + ([cls.CAN_FD_PADDING_PATTERN] * (padded_length - len(message)))

    def transmit(
        self,
        data: List[int],
        req_id: Optional[int] = None,
        extended: bool = False,
    ) -> None:
        """Pad and transmit a message through ITF connector. This method
        is also used as a substitute to the transmit method present in
        python-uds package.

        :param data: data to send.
        :param req_id: CAN message identifier. If not set use the one
            configured.
        :param extended: True if addressing mode is extended otherwise
            False.
        """
        req_id = req_id or self.req_id
        data = self._pad_message(data)
        self.channel._cc_send(msg=data, remote_id=req_id)

    def receive(self, timeout: float = 0) -> Optional[bytes]:
        """Receive a message through ITF connector. Called inside a thread,
        this method is a substitute to the reception method used in the
        python-uds package.

        :param timeout: Time to wait in second for a message to be received
        :return: the received message or None.
        """
        start_time = time.time()
        time_left = timeout

        while True:
            rcv_data = self.channel._cc_receive(timeout=timeout)
            msg, arbitration_id = rcv_data.get("msg"), rcv_data.get("remote_id")
            if msg is not None and arbitration_id == self.res_id:
                return msg

            time_left = timeout - (time.time() - start_time)

            if time_left < 0:
                return None

    def send_response(self, response_data: List[int]) -> None:
        """Encode and transmit a UDS response.

        :param response_data: the UDS response to send.
        """
        to_send = self.uds_config.tp.encode_isotp(response_data, use_external_snd_rcv_functions=True)
        if to_send is not None:
            self.transmit(to_send)

    @staticmethod
    def encode_stmin(stmin: float) -> int:
        """Encode the provided minimum separation time according to the ISO TP
        specification.

        :param stmin: minimum separation time in ms.
        :raises ValueError: if the provided value is not valid.
        :return: the encoded STmin to be sent in a flow control frame.
        """
        if stmin == 0:
            return stmin
        elif 1 <= stmin <= 127:
            # 1 - 127 ms -> 0x01 - 0x7F
            return int(stmin)
        elif 0.1 <= stmin <= 0.9:
            # 0.1 - 0.9 ms -> 0xF1 - 0xF9
            return 0xF0 + int(stmin * 10)
        else:
            raise ValueError(
                f"Invalid minimum Separation Time {stmin}ms. " "Acceptable values are between 0.1ms and 127ms."
            )

    def send_flow_control(self, flow_status: int = 0, block_size: int = 0, stmin: float = 0) -> None:
        """Send an ISO TP flow control frame to the client.

        :param flow_status: status of the flow control, defaults to 0
            (continue to send).
        :param block_size: size of the data block to send, defaults to 0
            (infinitely large).
        :param stmin: minimum separation time between 2 consecutive frames
            in ms, defaults to 0 ms.
        """
        flow_control_frame = [
            (0x30 + flow_status),
            block_size,
            self.encode_stmin(stmin),
        ]
        self.transmit(flow_control_frame)

    def register_callback(
        self,
        request: Union[int, List[int], UdsCallback, OdxRequestConfigDict],
        response: Optional[Union[int, List[int], Dict[str, str]]] = None,
        response_data: Optional[Union[int, bytes]] = None,
        data_length: Optional[int] = None,
        callback: Optional[Callable] = None,
    ) -> None:
        """Register an automatic response to send if the specified request is received
        from the client.

        The callback is stored inside the callbacks dictionary under the format
        {"0x2EC4": UdsCallback()}, where the keys are case-sensitive and
        correspond to the registered requests.
        If the callback is ODX based, a second key using Service.Parameter
        ("ReadByDataIdentifier.SoftwareVersion") key is registered.

        :param request: UDS request to be responded to.
        :param response: full UDS response to send. If not set, respond with a basic
            positive response with the specified response_data.
        :param response_data: UDS data to send. If not set, respond with a basic
            positive response containing no data.
        :param data_length: optional length of the data to send if it is supposed
            to have a fixed length (zero-padded).
        :param callback: custom callback to register
        """
        # handle odx based callbacks
        if isinstance(request, dict) or isinstance(response, dict):
            odx_param = self._get_odx_callback_param(request, response)
            request = self._create_callback_from_odx(request, response, response_data, data_length, callback)
            odx_key = f"{IsoServices(request.request[0]).name}.{odx_param}"
            self.callbacks[odx_key] = request
        elif isinstance(request, UdsCallback) and (
            isinstance(request.request, dict) or isinstance(request.response, dict)
        ):
            odx_param = self._get_odx_callback_param(request.request, request.response)
            request = self._create_callback_from_odx(
                request.request,
                request.response,
                request.response_data,
                request.data_length,
                request.callback,
            )
            odx_key = f"{IsoServices(request.request[0]).name}.{odx_param}"
            self.callbacks[odx_key] = request

        callback = (
            request
            if isinstance(request, UdsCallback)
            else UdsCallback(
                request=request,
                response=response,
                response_data=response_data,
                data_length=data_length,
                callback=callback,
            )
        )
        self.callbacks[self.format_data(callback.request)] = callback

    def unregister_callback(self, request: Union[str, int, List[int]]) -> None:
        """Unregister previously registered callback.

        The callback is stored inside the callbacks dictionary under the format
        {"0x2E01": UdsCallback()}, where the keys are case-sensitive and
        correspond to the registered requests.
        If more than one key references a callback, all are deleted.

        :param request: request for which the callback was registered as a
            string ("0x2E01"), an integer (0x2e01) or a list ([0x2e, 0x01]).
            If the callback is ODX based you can also use the readable key as str
            ("ReadByDataIdentifier.SoftwareVersion").
        """
        if isinstance(request, int):
            request = list(UdsCallback.int_to_bytes(request))
        if isinstance(request, list):
            request = self.format_data(request)
        with self._callback_lock:
            try:
                callback = self._callbacks.pop(request)
                for key, value in list(self._callbacks.items()):
                    if value == callback:
                        self._callbacks.pop(key)
            except KeyError as e:
                log.error(f"Could not unregister callback {e}: no such callback registered.")

    def _receive_message(self, timeout_in_s: float) -> None:
        """Reception method called by the auxiliary thread. This method received
        data and triggers the registered callbacks according to the received data.

        :param timeout_in_s: timeout on reception.
        """
        rcv_data = self.channel.cc_receive(timeout_in_s)
        msg, arbitration_id = rcv_data.get("msg"), rcv_data.get("remote_id")
        if msg is not None and arbitration_id == self.res_id:
            try:
                uds_data = self.uds_config.tp.decode_isotp(received_data=msg, use_external_snd_rcv_functions=True)
                if log.isEnabledFor(logging.getLogger().level):
                    log.internal_debug(
                        "Received ISO TP data: 0x%s || UDS data: %s",
                        msg.hex(),
                        self.format_data(uds_data),
                    )
            except Exception as e:
                # avoid timeouts that would break the thread
                log.exception(e)
                return

            self._dispatch_callback(uds_data)

    def _dispatch_callback(self, received_uds_data: List[int]) -> None:
        """Verify if the received UDS request has an associated response
        registered by a callback and send it.

        :param received_uds_data: received UDS request from the client.
        """
        for callback in self.callbacks.values():
            # match on the registered request instead of the entire received request
            if callback.request == received_uds_data[: len(callback.request)]:
                callback_to_execute = callback
                break
        else:
            log.internal_warning(f"Unregistered request received: {self.format_data(received_uds_data)}")
            return

        callback_to_execute(received_uds_data, self)
        return

    def _format_uds_data(self, coded_values: List[int]) -> List[int]:
        """Format a list of coded values into a list of bytes, needed for odx configured callbacks

        :param coded_values: coded values of a odx request
        :return: correctly formatted list of bytes for uds request
        """
        uds_data = []
        for value in coded_values:
            int_bytes = list(UdsCallback.int_to_bytes(value))
            uds_data.extend(int_bytes)
        return uds_data

    def _create_callback_from_odx(
        self,
        request: Dict,
        response: Optional[Union[int, List[int], Dict[str, str]]] = None,
        response_data: Optional[Union[int, bytes]] = None,
        data_length: Optional[int] = None,
        callback: Optional[Callable] = None,
    ) -> UdsCallback:
        """Register a UdsCallback from a dictionary containing ODX keywords

        :param request: contains the ODX data necessary to create a Uds request
        :param response: full UDS response to send. If not set, respond with a basic
            positive response with the specified response_data. Accepts ODX based dictionary
        :param response_data: UDS data to send. If not set, respond with a basic
            positive response containing no data.
        :param data_length: optional length of the data to send if it is supposed
            to have a fixed length (zero-padded).
        :param callback: custom callback to register
        :return: UdsCallback with request and response parsed from odx
        """
        if isinstance(request, dict):
            coded_values = self.odx_parser.get_coded_values(
                request["data"]["parameter"],
                request["service"],
                self.odx_parser.RefType.REQUEST,
            )
            uds_request = self._format_uds_data(coded_values)
            if uds_request[0] != request["service"]:
                log.error(f"Given SID {request['service']} does not match parsed SID {uds_request[0]}")
                raise ValueError(f"Given SID {request['service']} does not match parsed SID {uds_request[0]}")
        else:
            uds_request = request

        if isinstance(response, dict):
            # implementation for single values
            key, data = list(response.items())[0]
            if key.lower() == "negative":
                # create negative response: Negative response SID, request SID, NRC
                full_uds_response = [
                    UdsResponse.NEGATIVE_RESPONSE_SID,
                    uds_request[0],
                    data,
                ]
            else:
                # create positive response by parsing odx and adding data from dict
                coded_response_values = self.odx_parser.get_coded_values(
                    key,
                    uds_request[0] + 0x40,
                    self.odx_parser.RefType.POS_RESPONSE,
                )
                uds_response = self._format_uds_data(coded_response_values)
                payload = list(data.encode())
                full_uds_response = uds_response + payload
        else:
            full_uds_response = response

        callback = UdsCallback(
            uds_request,
            full_uds_response,
            response_data,
            data_length,
            callback,
        )
        log.internal_debug("Callback configured from odx: %s", callback)
        return callback

    def _get_odx_callback_param(
        self,
        request: Union[int, List[int], OdxRequestConfigDict],
        response: Optional[Union[int, List[int], Dict[str, str]]],
    ) -> str:
        """Used to create a readable odx based callback key

        :param request: contains the ODX data necessary to create a Uds request
        :param response: full UDS response to send. If not set, respond with a basic
            positive response with the specified response_data. Accepts ODX based dictionary
        :return: name of the parameter
        """
        logging.debug("--> req %s, res= %s", request, response)
        if isinstance(request, dict):
            key = request["data"]["parameter"]
            return key
        else:
            key = list(response.keys())[0]
            return key

    def _abort_command(self) -> None:
        """Not used, satisfy interface."""
        pass

    def _run_command(self, cmd_message, cmd_data=None) -> None:
        """Not used, satisfy interface."""
        pass



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
CAN Communication Channel using Vector hardware
***********************************************

:module: cc_vector_can

:synopsis: CChannel implementation for CAN(fd) using Vector API from python-can

.. currentmodule:: cc_vector_can

"""

import logging
from typing import Dict, Optional, Union

try:
    import can
    import can.bus
    import can.interfaces.vector
    from can.interfaces.vector.canlib import get_channel_configs
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[can]'")


from pykiso import CChannel, Message

MessageType = Union[Message, bytes]

log = logging.getLogger(__name__)


class CCVectorCan(CChannel):
    """CAN FD channel-adapter."""

    def __init__(
        self,
        bustype: str = "vector",
        poll_interval: float = 0.01,
        rx_queue_size: int = 524288,
        serial: int = None,
        channel: int = 3,
        bitrate: int = 500000,
        data_bitrate: int = 2000000,
        fd: bool = True,
        enable_brs: bool = False,
        app_name: str = None,
        can_filters: list = None,
        is_extended_id: bool = False,
        **kwargs,
    ):
        """Initialize can channel settings.

        :param bustype: python-can interface modules used
        :param poll_interval: Poll interval in seconds.
        :param rx_queue_size: Number of messages in receive queue
        :param serial: Vector Box's serial number. Can be replaced by the
            "AUTO" flag to trigger the Vector Box automatic detection.
        :param channel: The channel indexes to create this bus with
        :param bitrate: Bitrate in bits/s.
        :param app_name: Name of application in Hardware Config.
            If set to None, the channel should be a global channel index.
        :param data_bitrate: Which bitrate to use for data phase in CAN FD.
        :param fd: If CAN-FD frames should be supported.
        :param enable_brs: sets the bitrate_switch flag to use higher transmission speed
        :param can_filters: A iterable of dictionaries each containing
            a “can_id”, a “can_mask”, and an optional “extended” key.
        :param is_extended_id: This flag controls the size of the arbitration_id field.

        """
        super().__init__(**kwargs)
        self.bustype = bustype
        self.poll_interval = poll_interval
        self.rx_queue_size = rx_queue_size
        if str(serial).upper() == "AUTO":
            self.serial = detect_serial_number()
        else:
            self.serial = serial if not isinstance(serial, str) else int(serial)
        self.channel = channel
        self.app_name = app_name
        self.bitrate = bitrate
        self.data_bitrate = data_bitrate
        self.is_extended_id = is_extended_id
        self.fd = fd
        self.enable_brs = enable_brs
        self.can_filters = can_filters
        self.remote_id = None
        self.bus = None
        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6

        if self.enable_brs and not self.fd:
            log.internal_warning("Bitrate switch will have no effect because option is_fd is set to false.")

    def _cc_open(self) -> None:
        """Open a can bus channel and set filters for reception."""
        log.internal_info(f"CAN bus channel open: {self.channel}")
        self.bus = can.interface.Bus(
            bustype=self.bustype,
            poll_interval=self.poll_interval,
            rx_queue_size=self.rx_queue_size,
            serial=self.serial,
            app_name=self.app_name,
            channel=self.channel,
            bitrate=self.bitrate,
            data_bitrate=self.data_bitrate,
            fd=self.fd,
            bitrate_switch=self.enable_brs,
            can_filters=self.can_filters,
        )

    def _cc_close(self) -> None:
        """Close the current can bus channel."""
        log.internal_info(f"CAN bus channel closed: {self.channel}")
        self.bus.shutdown()
        self.bus = None

    def _cc_send(self, msg, remote_id: Optional[int] = None, **kwargs) -> None:
        """Send a CAN message at the configured id.

        If remote_id parameter is not given take configured ones.

        :param msg: data to send
        :param remote_id: destination can id used
        :param kwargs: named arguments

        """
        remote_id = remote_id or self.remote_id

        can_msg = can.Message(
            arbitration_id=remote_id,
            data=msg,
            is_extended_id=self.is_extended_id,
            is_fd=self.fd,
            bitrate_switch=self.enable_brs,
        )
        self.bus.send(can_msg)

        log.internal_debug(f"sent CAN Message: {can_msg}")

    def _cc_receive(self, timeout=0.0001) -> Dict[str, Union[MessageType, int]]:
        """Receive a can message using configured filters.

        :param timeout: timeout applied on reception

        :return: the received data and the source can id
        """
        try:  # Catch bus errors & rcv.data errors when no messages where received
            received_msg = self.bus.recv(timeout=timeout or self.timeout)

            if received_msg is not None:
                frame_id = received_msg.arbitration_id
                payload = received_msg.data
                timestamp = received_msg.timestamp

                log.internal_debug(f"received CAN Message: {frame_id}, {payload}")

                return {
                    "msg": payload,
                    "remote_id": frame_id,
                    "timestamp": timestamp,
                }
            else:
                return {"msg": None}
        except BaseException:
            log.exception(f"encountered error while receiving message via {self}")
            return {"msg": None}


def detect_serial_number() -> int:
    """Provide the serial number of the currently available Vector Box to be used.

    If several Vector Boxes are detected, the one with the lowest serial number is selected.
    If no Vector Box is connected, a ConnectionRefused error is thrown.

    :return: the Vector Box serial number
    :raises ConnectionRefusedError: raised if no Vector box is currently available
    """
    # Get all channels configuration
    channel_configs = get_channel_configs()
    # Getting all serial numbers
    serial_numbers = set()
    for channel_config in channel_configs:
        serial_number = channel_config.serial_number
        if serial_number != 0:
            serial_numbers.add(channel_config.serial_number)
    if serial_numbers:
        # if several devices are discovered, the first Vector Box is chosen
        serial_number = min(serial_numbers)
        log.internal_info(f"Using Vector Box with serial number {serial_number}")
        return serial_number
    else:
        raise ConnectionRefusedError("No Vector box is currently available")



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Interface Definition for Connectors, CChannels and Flasher
**********************************************************

:module: connector

:synopsis: Interface for a channel

.. currentmodule:: connector


"""
import abc
import logging
import pathlib
import threading
from typing import Dict, Optional

from .types import MsgType, PathType

log = logging.getLogger(__name__)


class Connector(abc.ABC):
    """Abstract interface for all connectors to inherit from.

    Defines hooks for opening and closing the connector and
    also defines a contextmanager interface."""

    def __init__(self, name: str = None):
        """Constructor.

        :param name: alias for the connector, used for ``repr`` and logging.
        """
        super().__init__()
        self.name = name

    def __repr__(self):
        name = self.name
        repr_ = super().__repr__()
        if name:
            repr_ = repr_[:1] + f"{name} is " + repr_[1:]
        return repr_

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, typ, value, traceback):
        self.close()
        if value is not None:
            raise value

    @abc.abstractmethod
    def open(self):
        """Initialise the Connector."""
        pass

    @abc.abstractmethod
    def close(self):
        """Close the connector, freeing resources."""
        pass


class CChannel(Connector):
    """Abstract class for coordination channel."""

    def __init__(self, processing=False, auto_open: bool = False, **kwargs: dict) -> None:
        """Constructor.

        :param processing: deprecated, will not be taken into account.
        :param auto_open: determine if the channel is automatically open.
        """
        super().__init__(**kwargs)

        self._lock_tx = threading.RLock()
        self._lock_rx = threading.RLock()
        self._lock = threading.Lock()
        self.auto_open = auto_open

    def open(self) -> None:
        """Open a thread-safe channel."""
        with self._lock:
            self._cc_open()

    def close(self) -> None:
        """Close a thread-safe channel."""
        with self._lock:
            self._cc_close()

    def shutdown(self) -> None:
        """Uninitialize channel. Will be called at the end of the test session."""
        pass

    def cc_send(self, msg: MsgType, *args, **kwargs) -> None:
        """Send a thread-safe message on the channel and wait for an acknowledgement.

        :param msg: message to send
        :param kwargs: named arguments
        """
        if ("raw" in kwargs) or args:
            log.internal_warning("Use of 'raw' keyword argument is deprecated. It won't be passed to '_cc_send'.")
        with self._lock_tx:
            self._cc_send(msg=msg, **kwargs)

    def cc_receive(self, timeout: float = 0.1, *args, **kwargs) -> Dict[str, Optional[bytes]]:
        """Read a thread-safe message on the channel and send an acknowledgement.

        :param timeout: time in second to wait for reading a message
        :param kwargs: named arguments

        :return: the received message
        """
        if ("raw" in kwargs) or args:
            log.internal_warning("Use of 'raw' keyword argument is deprecated. It won't be passed to '_cc_receive'.")
        with self._lock_rx:
            return self._cc_receive(timeout=timeout, **kwargs)

    @abc.abstractmethod
    def _cc_open(self) -> None:
        """Open the channel."""
        pass

    @abc.abstractmethod
    def _cc_close(self) -> None:
        """Close the channel."""
        pass

    @abc.abstractmethod
    def _cc_send(self, msg: MsgType, **kwargs) -> None:
        """Sends the message on the channel.

        :param msg: Message to send out
        :param kwargs: named arguments
        """
        pass

    @abc.abstractmethod
    def _cc_receive(self, timeout: float, **kwargs) -> Dict[str, Optional[bytes]]:
        """How to receive something from the channel.

        :param timeout: Time to wait in second for a message to be received
        :param kwargs: named arguments
        :return: dictionary containing the received bytes if successful, otherwise None
        """
        pass


class Flasher(Connector):
    """Interface for devices that can flash firmware on our targets."""

    def __init__(self, binary: PathType = None, **kwargs):
        """Constructor.

        :param binary: binary firmware file

        :raise ValueError: if binary doesn't exist or is not a file
        :raise TypeError: if given binary is None
        """
        super().__init__(**kwargs)
        if binary is not None:
            binary = pathlib.Path(binary).resolve()  # force an absolute path
            if binary.exists() and binary.is_file():
                self.binary = binary
            else:
                raise ValueError(f"'binary' must be a path-like object to an existing file (got {binary})")
        else:
            raise TypeError("'binary' must be a path-like object, not None")

    @abc.abstractmethod
    def flash(self):
        """Flash firmware on the target."""
        pass



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Process Channel
***************

:module: cc_process

:synopsis: CChannel implementation for process execution.

The CCProcess channel provides functionality to start a process and
to communicate with it.

.. currentmodule:: cc_process

"""

import logging
import queue
import subprocess
import threading
from dataclasses import dataclass
from typing import IO, ByteString, List, Optional, Union

from pykiso.connector import CChannel

log = logging.getLogger(__name__)

MessageType = Union[str, ByteString]


class CCProcessError(BaseException):
    ...


@dataclass
class ProcessMessage:
    """Holds the data that is read from the process"""

    # Stream name: stdout or stderr
    stream: str
    # Data as bytes or string
    data: Union[str, bytes]


@dataclass
class ProcessExit:
    """Contains information about process exit"""

    exit_code: int


class CCProcess(CChannel):
    """Channel to run processes"""

    def __init__(
        self,
        shell: bool = False,
        pipe_stderr: bool = False,
        pipe_stdout: bool = True,
        pipe_stdin: bool = False,
        text: bool = True,
        cwd: Optional[str] = None,
        env: Optional[str] = None,
        encoding: Optional[str] = None,
        executable: Optional[str] = None,
        args: List[str] = [],
        **kwargs,
    ):
        """Initialize a process

        :param shell: Start process through shell
        :param pipe_stderr: Pipe stderr for reading with this connector
        :param pipe_stdout: Pipe stdout for reading with this connector
        :param pipe_stdin:  Pipe stdin for writing with this connector
        :param text: Read/write stdout, stdin, stderr in binary mode
        :param cwd: The current working directory for the new process
        :param env: Environment variables for the new process
        :param encoding: Encoding to use in text mode
        :param executable: The path of the executable for the process
        :param args: Process arguments

        """
        super().__init__(**kwargs)
        self._shell = shell
        self._pipe_stderr = pipe_stderr
        self._pipe_stdout = pipe_stdout
        self._pipe_stdin = pipe_stdin
        self._encoding = encoding
        self._executable = executable
        self._args = args
        self._text = text
        self._cwd = cwd
        self._env = env
        self._process: Optional[subprocess.Popen] = None
        self._queue_in: Optional[queue.Queue[Union[ProcessMessage, ProcessExit]]] = None
        self._stdout_thread: Optional[threading.Thread] = None
        self._stderr_thread: Optional[threading.Thread] = None
        self._lock = threading.Lock()
        self._finished_threads_count = 0
        # Buffer for messages that where read from the process but not yet returned by _cc_receive
        self._buffer: List[Union[ProcessMessage, ProcessExit]] = []

    def start(
        self,
        executable: Optional[str] = None,
        args: Optional[List[str]] = None,
    ):
        """Start a process

        :param executable: The executable path. Default to path specified in yaml if not given.
        :param args: The process arguments. Default to arguments specified in yaml if not given.

        :raises CCProcessError: Process is already running
        """
        if self._process is not None and self._process.returncode is None:
            raise CCProcessError(f"Process is already running: {self._executable}")

        self._cleanup()
        self._finished_threads_count = 0
        self._queue_in = queue.Queue()

        self._process = subprocess.Popen(
            ([executable] if executable is not None else [self._executable])
            + (args if args is not None else self._args),
            stderr=subprocess.PIPE if self._pipe_stderr else None,
            stdout=subprocess.PIPE if self._pipe_stdout else None,
            stdin=subprocess.PIPE if self._pipe_stdin else None,
            shell=self._shell,
            text=self._text,
            encoding=self._encoding,
            cwd=self._cwd,
            env=self._env,
        )  # nosec B602 Since we only provide an interface to the user to popen, we accept the risk of a vulnerability to various shell injection attacks.

        if self._pipe_stdout:
            self._stdout_thread = self._start_read_thread(self._process.stdout, "stdout")
        if self._pipe_stderr:
            self._stderr_thread = self._start_read_thread(self._process.stderr, "stderr")

    def _start_read_thread(self, stream: IO, name: str) -> threading.Thread:
        """Start a read thread

        :param stream: The stream to read from
        :param name: The name of the stream

        :return: The thread object
        """
        thread = threading.Thread(
            name=f"cc_process_{name}",
            target=self._read_thread,
            args=(stream, name),
        )
        thread.start()
        return thread

    def _read_thread(self, stream: IO, name: str) -> None:
        """Thread for reading data from stdout or stderr

        :param stream: The stream to read from
        :param name: The name of the stream
        """
        try:
            while True:
                if self._text:
                    data = stream.readline()
                else:
                    data = stream.read(1)
                if len(data) == 0:
                    break
                self._queue_in.put(ProcessMessage(name, data))
        finally:
            with self._lock:
                self._finished_threads_count += 1
                if self._finished_threads_count == int(self._pipe_stdout) + int(self._pipe_stderr):
                    # ProcessExit marks the termination of all read threads
                    self._queue_in.put(ProcessExit(self._process.wait()))

    def _cc_close(self) -> None:
        """Close the channel."""
        self._cleanup()

    def _cc_send(self, msg: MessageType, **kwargs) -> None:
        """Execute process commands or write data to stdin

        :param msg: data to send

        :raises CCProcessError: Stdin pipe is not enabled

        """
        if isinstance(msg, dict) and msg.get("command") == "start":
            self.start(msg.get("executable"), msg.get("args"))
        elif self._pipe_stdin:
            if self._process is None:
                raise CCProcessError("Process is not running.")
            log.internal_debug(f"write stdin: {msg}")
            self._process.stdin.write(msg)
            self._process.stdin.flush()
        else:
            raise CCProcessError("Can not send to stdin because pipe is not enabled.")

    def _cleanup(self) -> None:
        """Cleanup threads and process objects"""
        if self._process is not None:
            # Terminate the process if still running
            self._process.terminate()
            try:
                self._process.wait(5)
            except subprocess.TimeoutExpired:
                log.internal_warning(f"Process {self._executable} could not be terminated")
                self._process.kill()
        # Wait for the threads to finish
        if self._stdout_thread is not None:
            self._stdout_thread.join()
            self._stdout_thread = None
        if self._stderr_thread is not None:
            self._stderr_thread.join()
            self._stderr_thread = None
        if self._queue_in is not None:
            self._queue_in = None

    def _cc_open(self) -> None:
        """Implement abstract method"""
        pass

    def _read_existing(self) -> Optional[ProcessMessage]:
        """Read buffered messages that where already received from the process.
        Messages from the same stream are combined.
        This is only used in binary mode.

        :return: Existing messages
        """
        messages = self._buffer

        # Get all messages from the process that are available
        while not self._queue_in.empty():
            messages.append(self._queue_in.get_nowait())
        i = 1
        # Find messages from the same stream(first entry in the tuple) as the first message
        while (
            i < len(messages)
            and not isinstance(messages[0], ProcessExit)
            and not isinstance(messages[i], ProcessExit)
            and messages[0].stream == messages[i].stream
        ):
            i += 1

        # Save the remaining messages for next time
        messages, self._buffer = messages[:i], messages[i:]

        # Process only messages from the same stream
        messages = messages[:i]
        if len(messages) == 0:
            return None

        if isinstance(messages[0], ProcessExit):
            return messages[0]

        # Join messages
        return ProcessMessage(messages[0].stream, b"".join([x.data for x in messages]))

    @staticmethod
    def _create_message_dict(msg: Union[ProcessMessage, ProcessExit]) -> dict:
        """Create a dict from an entry in the process queue

        :param msg: The message to convert

        :return: The dictionary
        """
        if isinstance(msg, ProcessMessage):
            ret = {"msg": {msg.stream: msg.data}}
        elif isinstance(msg, ProcessExit):
            ret = {"msg": {"exit": msg.exit_code}}
        return ret

    def _cc_receive(self, timeout: float = 0.0001) -> MessageType:
        """Receive messages

        :param timeout: Time to wait in seconds for a message to be received
        :param size: unused

        return The received message
        """
        if self._queue_in is None:
            return {"msg": None}

        # Get message from the queue
        try:
            read = self._queue_in.get(True, timeout)
        except queue.Empty:
            # Queue is empty, but there might be previously received messages when in binary mode
            existing = None if self._text else self._read_existing()
            if existing is not None:
                ret = CCProcess._create_message_dict(existing)
            else:
                ret = {"msg": None}
            return ret

        if not isinstance(read, ProcessExit):
            # A message was received
            if self._text:
                # Just return that message when in text mode
                ret = CCProcess._create_message_dict(read)
            else:
                # Add message to the buffer and join messages for binary mode
                self._buffer.append(read)
                existing = self._read_existing()
                ret = CCProcess._create_message_dict(existing)
            return ret
        else:
            # Process has exited
            self._cleanup()
            return CCProcess._create_message_dict(read)



##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Can Auxiliary
*************

:module: can_auxiliary

:synopsis: Auxiliary enables you to read write can messages which are defined in a dbc file.

The goal of this module is to be able to receive and send signals which are defined in dbc file.

.. currentmodule:: can_auxiliary
"""
from __future__ import annotations

import functools
import logging
import threading
import time
from contextlib import ContextDecorator
from copy import deepcopy
from queue import Empty, Queue
from typing import Any, Optional

from pykiso import Message
from pykiso.auxiliary import AuxiliaryInterface, close_connector, open_connector
from pykiso.connector import CChannel

from .can_message import CanMessage
from .can_parser import CanMessageParser

log = logging.getLogger(__name__)


class _collect_messages(ContextDecorator):
    """Context manager and decorator for the can auxiliary
    allowing messages collection (putting them in a list of the auxiliary).
    """

    def __init__(self, can_aux: CanAuxiliary):
        """Constructor used to inherit some of can auxiliary features."""
        self.can_aux = can_aux

    def __enter__(self):
        """Start adding the message received in a list"""
        log.internal_debug("Start collecting received can messages.")
        self.can_aux._messages_collected = []
        self.can_aux._collect_msg.set()

    def __exit__(self, *exc):
        """Stop adding message received in a list."""
        log.internal_debug("Stop collecting received can messages.")
        self.can_aux._collect_msg.clear()


class CanAuxiliary(AuxiliaryInterface):
    """Auxiliary is used for reading and writing can messages defined in dbc file"""

    def __init__(self, com: CChannel, dbc_file: str, **kwargs):
        """Constructor.

        :param com: CChannel that supports raw communication over CAN
        :param dbc_file: dbc file that provides the can messages structure
        """
        self.tx_task_on = False
        super().__init__(is_proxy_capable=True, tx_task_on=True, rx_task_on=True, **kwargs)

        self.recv_can_messages = {}
        self.channel = com
        path_to_dbc_file = dbc_file
        self.parser = CanMessageParser(path_to_dbc_file)
        self.can_messages = {}
        # Variables to manage the collection of can messages with a context-manager
        self._collect_msg = threading.Event()
        self._messages_collected = []
        self.collect_messages = functools.partial(_collect_messages, can_aux=self)

    @open_connector
    def _create_auxiliary_instance(self) -> bool:
        """Open the connector communication.

        :return: True if the channel is correctly opened otherwise False
        """
        log.internal_info("Auxiliary instance created")
        return True

    @close_connector
    def _delete_auxiliary_instance(self) -> bool:
        """Close the connector communication.

        :return: always True
        """
        log.internal_info("Auxiliary instance deleted")
        return True

    def _receive_message(self, timeout_in_s: float = 0.1) -> None:
        """Trigger connector reception method and put received messages
        in the queue.

        :param timeout_in_s: maximum time in second to wait for a response
        """
        try:
            rcv_data = self.channel.cc_receive(timeout=timeout_in_s)
            if rcv_data.get("msg") is not None:
                log.internal_debug(f"received message '{rcv_data}' from {self.channel}")
                message_name = self.parser.dbc.get_message_by_frame_id(rcv_data["remote_id"]).name
                message_signals = self.parser.decode(rcv_data["msg"], rcv_data["remote_id"])
                message_timestamp = float(rcv_data.get("timestamp", 0))
                old_can_message = self.can_messages.get(message_name, None)
                if old_can_message is None:
                    self.can_messages[message_name] = Queue(maxsize=1)
                if not self.can_messages[message_name].empty():
                    self.can_messages[message_name].get_nowait()
                can_msg = CanMessage(message_name, message_signals, message_timestamp)
                self.can_messages[message_name].put(can_msg)
                if self._collect_msg.is_set():
                    self._messages_collected.append(can_msg)
        except KeyError:
            log.exception("Specific message signal is not found in message")
            pass
        except Exception:
            log.exception(f"encountered error while receiving message via {self.channel}")
            pass

    def get_last_message(self, message_name: str) -> Optional[CanMessage]:
        """Get the last message which has been received on the bus.

        :param message_name: name of the message, you want to get

        :return: last can massage or return none if the message, or return none if message is not occur
        """

        can_msg_queue = self.can_messages.get(message_name, None)
        if can_msg_queue is not None and not can_msg_queue.empty():
            msg = can_msg_queue.get_nowait()
            self.can_messages[message_name].put_nowait(msg)
            return msg
        return None

    def get_last_signal(self, message_name: str, signal_name: str) -> Optional[Any]:
        """Get the last message which has been received on the bus.

        :param message_name: name of the message, you want to get
        :param signal_name: name of the signal, you want to get

        :return: last can massage or return none if the message or return none if message or signal is not occur
        """
        can_msg_queue = self.can_messages.get(message_name, None)
        if can_msg_queue is not None and not can_msg_queue.empty():
            last_can_message = can_msg_queue.get_nowait()
            if last_can_message is not None:
                self.can_messages[message_name].put_nowait(last_can_message)
                return last_can_message.signals.get(signal_name, None)

        return None

    def get_collected_messages(self) -> list[CanMessage]:
        """Get all the messages collected with the context manager

        :return: a list with all the collected messages
        """
        return deepcopy(self._messages_collected)

    def wait_for_message(self, message_name: str, timeout: float = 0.2) -> dict[str, any]:
        """Get the last message with certain timeout in seconds.

        :param message_name: name of the message to receive
        :param timeout: time to wait till a message receives in seconds

        :return: list of last can messages or None if no messages for this component
        """

        can_msg_queue = self.can_messages.get(message_name, None)
        old_value = None
        if can_msg_queue is None:
            self.can_messages[message_name] = Queue(maxsize=1)

        if not self.can_messages[message_name].empty():
            old_value = self.can_messages[message_name].get()

        try:
            new_msg = self.can_messages[message_name].get(timeout=timeout)
            self.can_messages[message_name].put_nowait(new_msg)
            return new_msg
        except Empty:
            if old_value is not None:
                self.can_messages[message_name].put_nowait(old_value)
            return None

    def wait_to_match_message_with_signals(
        self,
        message_name: str,
        expected_signals: dict[str, any],
        timeout: float = 0.2,
    ) -> dict[str, any]:
        """Get first message which matches the patter of signals.

        :param message_name: name of the message to receive
        :param expected_signal: list of expected signals to match with message
        :param timeout: time to wait till a message receives in seconds

        :return: list of last can messages or None if no messages for this component
        """

        t1 = time.perf_counter()
        message_to_return = None
        is_msg_matched = False
        while time.perf_counter() - t1 < timeout and not is_msg_matched:
            expected_signals_names = expected_signals.keys()
            last_msg_queue = self.can_messages.get(message_name, None)
            if last_msg_queue is None:
                continue
            try:
                last_can_msg = last_msg_queue.get_nowait()
            except Empty:
                continue
            is_msg_matched = True
            for signal_name in expected_signals_names:
                if last_can_msg.signals[signal_name] != expected_signals[signal_name]:
                    is_msg_matched = False
                    break
            if is_msg_matched:
                message_to_return = last_can_msg
                self.can_messages[message_name].put_nowait(last_can_msg)
                break

        return message_to_return

    def send_message(self, message: str, signals: dict[str, Any]) -> bool:
        """Send one message, the message need to be defined in the dbc file.

        :param message: name of the message to send.
        :param signals: dict of the signals of the message and their value.

        :return: True or False if the message has been successfully send or not.
        """

        for signal, value in signals.items():
            if isinstance(value, str):
                signals[signal] = int.from_bytes(value.encode("utf8"), byteorder="big")
        try:
            message = self.parser.dbc.get_message_by_name(message)
        except KeyError:
            raise ValueError(f"{message} is not a message defined in the DBC file.")
        msg_to_send = self.parser.encode(message, signals)
        self.channel.cc_send(msg_to_send[0], remote_id=msg_to_send[1])

    def _run_command(self, cmd_message: str, cmd_data: bytes = None) -> bool:
        """Run the corresponding command.

        :param cmd_message: command type
        :param cmd_data: payload data to send over CChannel

        :return: True if command is executed otherwise False
        """
        if cmd_message == "send":
            try:
                self.channel.cc_send(msg=cmd_data)
            except Exception:
                log.exception(f"encountered error while sending message '{cmd_data}' to {self.channel}")
        elif isinstance(cmd_message, Message):
            log.internal_debug(f"ignored command '{cmd_message} in {self}'")
        else:
            log.internal_warning(f"received unknown command '{cmd_message} in {self}'")



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Communication Channel via UDP server
************************************

:module: cc_udp_server

:synopsis: basic UDP server

.. currentmodule:: cc_udp_server

.. warning:: if multiple clients are connected to this server,
    ensure that each client receives all necessary responses before
    receiving messages again. Otherwise the responses may be
    sent to the wrong client

"""
import logging
import socket
from typing import Dict, Optional, Union

from pykiso import Message, connector

log = logging.getLogger(__name__)


class CCUdpServer(connector.CChannel):
    """Connector channel used to set up an UDP server."""

    def __init__(self, dest_ip: str, dest_port: int, **kwargs):
        """Initialize attributes.

        :param dest_ip: destination port
        :param dest_port: destination port
        """
        super().__init__(**kwargs)
        self.dest_ip = dest_ip
        self.dest_port = dest_port
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.address = None
        self.max_msg_size = 256
        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6

    def _cc_open(self) -> None:
        """Bind UDP socket with configured port and IP address."""
        log.internal_info(f"UDP socket open at address: {self.address}")
        self.udp_socket.bind((self.dest_ip, self.dest_port))

    def _cc_close(self) -> None:
        """Close UDP socket."""
        log.internal_info(f"UDP socket closed at address: {self.address}")
        self.udp_socket.close()

    def _cc_send(self, msg: bytes, **kwargs) -> None:
        """Send back a UDP message to the previous sender.

        :param msg: message to sent, should be bytes
        :param kwargs: not used
        """
        log.internal_debug(f"UDP server send: {msg} at {self.address}")
        self.udp_socket.sendto(msg, self.address)

    def _cc_receive(self, timeout=0.0000001) -> Dict[str, Optional[bytes]]:
        """Read message from UDP socket.

        :param timeout: timeout applied on receive event

        :return: Message if successful, otherwise none
        """

        self.udp_socket.settimeout(timeout or self.timeout)

        try:
            msg_received, self.address = self.udp_socket.recvfrom(self.max_msg_size)
            log.internal_debug(f"UDP server receives: {msg_received} at {self.address}")
        # catch the errors linked to the socket timeout without blocking
        except BlockingIOError:
            log.internal_debug(f"encountered error while receiving message via {self}")
            return {"msg": None}
        except socket.timeout:
            log.internal_debug(f"encountered error while receiving message via {self}")
            return {"msg": None}
        except BaseException:
            log.exception(f"encountered error while receiving message via {self}")
            return {"msg": None}

        return {"msg": msg_received}



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Scenario
********

:module: scenario

:synopsis: base object used to create pre-defined virtual DUT scenario.

.. currentmodule:: scenario

.. warning:: Still under test
"""

from collections import UserList

from .response_templates import ResponseTemplates


class Scenario(UserList):
    """Container used to create pre-defined virtual DUT scenario."""

    pass


class TestScenario:
    """Encapsulate all possible test's scenarios."""

    @classmethod
    def handle_successful(cls) -> Scenario:
        """Return a scenario to handle a complete successful
        test case exchange(TEST CASE setup->run->teardown).

        :return: Scenario instance containing all steps
        """
        return Scenario(
            [
                ResponseTemplates.ack_with_report_ok,
                ResponseTemplates.ack_with_report_ok,
                ResponseTemplates.ack_with_report_ok,
            ]
        )

    @classmethod
    def handle_communication_lost(cls) -> Scenario:
        return Scenario([ResponseTemplates.ack])

    class VirtualTestCase:
        """Used to gather all virtual DUT test case scenarios based on
        their fixture level (setup, run, teardown).
        """

        class Setup:
            """Used to gather all possible scenarios linked to a test
            case setup execution.
            """

            @classmethod
            def handle_failed_report_setup(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with failed report at setup phase.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_nok,
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_ok,
                    ]
                )

            @classmethod
            def handle_not_implemented_report_setup(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with not implemented report at setup phase.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_not_implemented,
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_ok,
                    ]
                )

            @classmethod
            def handle_lost_communication_during_setup_ack(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with lost of communication during ACK to setup Command.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        None,
                        None,
                        ResponseTemplates.ack,
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_ok,
                    ]
                )

            @classmethod
            def handle_lost_communication_during_setup_report(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with lost of communication during report to setup Command.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack,
                        None,
                        None,
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_ok,
                    ]
                )

        class Run:
            """Used to gather all possible scenarios linked to a test case
            run execution.
            """

            @classmethod
            def handle_failed_report_run(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with failed report at run phase.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_nok,
                        ResponseTemplates.ack_with_report_ok,
                    ]
                )

            @classmethod
            def handle_successful_report_run_with_log(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with successful log and report at run phase.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_logs_and_report_ok,
                        ResponseTemplates.ack_with_report_ok,
                    ]
                )

            @classmethod
            def handle_failed_report_run_with_log(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with failed log and report at run phase.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_logs_and_report_nok,
                        ResponseTemplates.ack_with_report_ok,
                    ]
                )

            @classmethod
            def handle_not_implemented_report_run(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with not implemented report at run phase.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_not_implemented,
                        ResponseTemplates.ack_with_report_ok,
                    ]
                )

            @classmethod
            def handle_lost_communication_during_run_ack(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with lost of communication during ACK to run Command.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_ok,
                        None,
                        None,
                        ResponseTemplates.ack,
                        ResponseTemplates.ack_with_report_ok,
                    ]
                )

            @classmethod
            def handle_lost_communication_during_run_report(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with lost of communication during report to run Command.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack,
                        None,
                        None,
                        ResponseTemplates.ack_with_report_ok,
                    ]
                )

        class Teardown:
            """Used to gather all possible scenarios linked to a test case
            teardown execution.
            """

            @classmethod
            def handle_failed_report_teardown(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with failed report at teardown phase.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_nok,
                    ]
                )

            @classmethod
            def handle_not_implemented_report_teardown(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with not implemented report at teardown phase.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_not_implemented,
                    ]
                )

            @classmethod
            def handle_lost_communication_during_teardown_ack(cls) -> Scenario:
                """Return a scenario to handle a complete test case
                with lost of communication during ACK to teardown Command.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_ok,
                        None,
                        None,
                        ResponseTemplates.ack,
                    ]
                )

            @classmethod
            def handle_lost_communication_during_teardown_report(
                cls,
            ) -> Scenario:
                """Return a scenario to handle a complete test case
                with lost of communication during report to teardown Command.

                :return: Scenario instance containing all steps
                """
                return Scenario(
                    [
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack_with_report_ok,
                        ResponseTemplates.ack,
                        None,
                        None,
                    ]
                )

    class VirtualTestSuite:
        """Used to gather all virtual DUT test suite scenarios based on
        their fixture level (setup, teardown).
        """

        class Setup:
            """Used to gather all possible scenarios linked to a test suite
            setup execution.
            """

            @classmethod
            def handle_failed_report_setup(cls) -> Scenario:
                """Return a scenario to handle a test suite setup
                with report failed.

                :return: Scenario instance containing all steps
                """
                return Scenario([ResponseTemplates.ack_with_report_nok])

            @classmethod
            def handle_not_implemented_report_setup(cls) -> Scenario:
                """Return a scenario to handle a test suite setup
                with report not implemented.

                :return: Scenario instance containing all steps
                """
                return Scenario([ResponseTemplates.ack_with_report_not_implemented])

            @classmethod
            def handle_lost_communication_during_setup_ack(cls) -> Scenario:
                """Return a scenario to handle a lost of communication
                during ACK to setup command.

                :return: Scenario instance containing all steps
                """
                return Scenario([None, None, ResponseTemplates.ack])

            @classmethod
            def handle_lost_communication_during_setup_report(cls) -> Scenario:
                """Return a scenario to handle a lost of communication
                during report to setup command.

                :return: Scenario instance containing all steps
                """
                return Scenario([ResponseTemplates.ack, None, None])

        class Teardown:
            """Used to gather all possible scenarios linked to a test suite
            teardown execution.
            """

            @classmethod
            def handle_failed_report_teardown(cls) -> Scenario:
                """Return a scenario to handle a test suite teardown
                with report failed.

                :return: Scenario instance containing all steps
                """
                return Scenario([ResponseTemplates.ack_with_report_nok])

            @classmethod
            def handle_not_implemented_report_teardown(cls) -> Scenario:
                """Return a scenario to handle a test suite teardown
                with report not implemented.

                :return: Scenario instance containing all steps
                """
                return Scenario([ResponseTemplates.ack_with_report_not_implemented])

            @classmethod
            def handle_lost_communication_during_teardown_ack(cls) -> Scenario:
                """Return a scenario to handle a lost of communication
                during ACK to teardown command.

                :return: Scenario instance containing all steps
                """
                return Scenario([None, None, ResponseTemplates.ack])

            @classmethod
            def handle_lost_communication_during_teardown_report(
                cls,
            ) -> Scenario:
                """Return a scenario to handle a lost of communication
                during report to teardown command.

                :return: Scenario instance containing all steps
                """
                return Scenario([ResponseTemplates.ack, None, None])



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Fake Flasher Channel for testing
********************************

:module: cc_flasher_example

:synopsis: fake flasher implementation

.. warning: ONLY FOR TESTING PURPOSE!
"""

import logging

from pykiso import Flasher

log = logging.getLogger(__name__)


class FlasherExample(Flasher):
    """A Flasher adapter for testing purpose only."""

    def __init__(self, name: str, **kwargs) -> None:
        """Constructor.

        :param name: flasher's alias
        :param kwargs: named arguments
        """
        self.name = name

    def open(self) -> None:
        """Initialize the flasher."""
        log.internal_info("Initialize flasher")
        log.internal_info("Flasher ready!")

    def close(self) -> None:
        """Close flasher and free resources."""
        log.internal_info("Close flasher")
        log.internal_info("Free ressoucres!")

    def flash(self) -> None:
        """Fake a firmware update."""
        log.internal_info("flashing device")
        log.internal_info("flashing device successful")



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
ResponseTemplates
*****************

:module: response_templates

:synopsis: Used to create a set of predefined messages

.. currentmodule:: response_templates

.. warning:: Still under test
"""
import random
from typing import List

from pykiso import message
from pykiso.message import Message


class ResponseTemplates:
    """Used to create a set of predefined messages (ACK, NACK, REPORT
    ...)."""

    reasons_list = [
        "Component busy, could not run",
        "fatal error occur, could not run",
        "Overvoltage situation happen, could not run",
    ]
    log = []

    @classmethod
    def default(cls, msg: Message) -> Message:
        """handle default response, if not test case/suite run
        just return ACK message otherwise ACK + REPORT.

        :param msg: current received message

        :return: list of Message
        """

        if (
            msg.sub_type in message.MessageCommandType.__members__.values()
            and msg.sub_type != message.MessageCommandType.PING
            and msg.sub_type != message.MessageCommandType.ABORT
        ):
            return cls.ack_with_report_ok(msg)
        else:
            return cls.ack(msg)

    @classmethod
    def ack(cls, msg: Message) -> List[Message]:
        """Return an acknowledgment message.

        :param msg: current received message

        :return: list of Message
        """
        return [msg.generate_ack_message(message.MessageAckType.ACK)]

    @classmethod
    def ack_with_report_ok(cls, msg: Message) -> List[Message]:
        """Return an acknowledgment message and a report message with
        verdict pass.

        :param msg: current received message

        :return: list of Message
        """
        ack_msg = msg.generate_ack_message(message.MessageAckType.ACK)

        report_msg = message.Message(
            msg_type=message.MessageType.REPORT,
            sub_type=message.MessageReportType.TEST_PASS,
            test_suite=msg.test_suite,
            test_case=msg.test_case,
        )

        return [ack_msg, report_msg]

    @classmethod
    def ack_with_report_nok(cls, msg: Message) -> List[Message]:
        """Return an acknowledgment message and a report message with
        verdict failed + tlv part with failure reason.

        :param msg: current received message

        :return: list of Message
        """

        ack_run_cmd = msg.generate_ack_message(message.MessageAckType.ACK)

        report_msg = message.Message(
            msg_type=message.MessageType.REPORT,
            sub_type=message.MessageReportType.TEST_FAILED,
            test_suite=msg.test_suite,
            test_case=msg.test_case,
        )

        report_msg.tlv_dict = cls.get_random_reason()

        return [ack_run_cmd, report_msg]

    @classmethod
    def ack_with_logs_and_report_ok(cls, msg: Message) -> List[Message]:
        """Return an acknowledge message and log messages and
        report message with verdict pass.

         :param msg: current received message

         :return: list of Message
        """

        ack_msg = msg.generate_ack_message(message.MessageAckType.ACK)

        log_msg = message.Message(
            msg_type=message.MessageType.LOG,
            sub_type=message.MessageLogType.RESERVED,
            test_suite=msg.test_suite,
            test_case=msg.test_case,
        )

        report_msg = message.Message(
            msg_type=message.MessageType.REPORT,
            sub_type=message.MessageReportType.TEST_PASS,
            test_suite=msg.test_suite,
            test_case=msg.test_case,
        )

        return [ack_msg, log_msg, log_msg, report_msg]

    @classmethod
    def ack_with_logs_and_report_nok(cls, msg: Message) -> List[Message]:
        """Return an acknowledge message and log messages and
        report message with verdict failed + tlv part with failure reason.

         :param msg: current received message

         :return: list of Message
        """

        ack_msg = msg.generate_ack_message(message.MessageAckType.ACK)

        log_msg = message.Message(
            msg_type=message.MessageType.LOG,
            sub_type=message.MessageLogType.RESERVED,
            test_suite=msg.test_suite,
            test_case=msg.test_case,
        )

        report_msg = message.Message(
            msg_type=message.MessageType.REPORT,
            sub_type=message.MessageReportType.TEST_FAILED,
            test_suite=msg.test_suite,
            test_case=msg.test_case,
        )

        report_msg.tlv_dict = cls.get_random_reason()

        return [ack_msg, log_msg, log_msg, report_msg]

    @classmethod
    def get_random_reason(self) -> dict:
        """Return tlv dictionary containing a random reason from
        pre-defined reason list.

        :param msg: current received message

        :return: tlv dictionary with failure reason
        """
        return {message.TlvKnownTags.FAILURE_REASON: random.choice(ResponseTemplates.reasons_list)}

    @classmethod
    def nack_with_reason(cls, msg: Message) -> List[Message]:
        """Return a NACK message with a tlv part
        containing the failure reason.

        :param msg: current received message

        :return: list of Message
        """
        ack_msg = msg.generate_ack_message(message.MessageAckType.NACK)
        ack_msg.tlv_dict.update(cls.get_random_reason())

        return [ack_msg]

    @classmethod
    def ack_with_report_not_implemented(cls, msg: Message) -> List[Message]:
        """Return an acknowledge message and a report message with
        verdict test not implemented.

        :param msg: current received message

        :return: list of Message
        """

        ack_run_cmd = msg.generate_ack_message(message.MessageAckType.ACK)

        report_msg = message.Message(
            msg_type=message.MessageType.REPORT,
            sub_type=message.MessageReportType.TEST_NOT_IMPLEMENTED,
            test_suite=msg.test_suite,
            test_case=msg.test_case,
        )

        return [ack_run_cmd, report_msg]



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Virtual DUT simulation package
******************************

:module: simulated_auxiliary

:synopsis: provide a simple interface to simulate a device under test

This auxiliary can be used as a simulated version of a device under test.

The intention is to set up a pair of CChannels like a pipe, for example a
:class:`~pykiso.lib.connectors.cc_udp_server.CCUdpServer` and a :class:`~pykiso.lib.connectors.cc_udp.CCUdp`
bound to the same address.
One side of this pipe is then connected to this virtual auxiliary,
the other one to a *real* auxiliary.

The :class:`~pykiso.lib.auxiliary.simulated_auxiliary.simulated_auxiliary.SimulatedAuxiliary` will then receive messages
from the real auxiliary just like a proper TestApp on a DUT would and answer them according to a predefined
playbook.

Each predefined playbooks are linked with real auxiliary received messages, using test case and test suite ids
(see :mod:`~pykiso.lib.auxiliaries.simulated_auxiliary.simulation`).
A so called playbook, is a basic list of different :class:`~pykiso.message.Message` instances where the content
is adapted to the current context under test (simulate a communication lost, a test case run failure...).
(see :mod:`~pykiso.lib.auxiliaries.simulated_auxiliary.scenario`).
In order to increase playbook configuration flexibility, predefined and reusable responses are located
into :mod:`~pykiso.lib.auxiliaries.simulated_auxiliary.response_templates`.

.. currentmodule:: simulated_auxiliary

.. autosummary::

   pykiso.lib.auxiliaries.simulated_auxiliary.simulated_auxiliary
   pykiso.lib.auxiliaries.simulated_auxiliary.simulation
   pykiso.lib.auxiliaries.simulated_auxiliary.scenario
   pykiso.lib.auxiliaries.simulated_auxiliary.response_templates

.. automodule:: pykiso.lib.auxiliaries.simulated_auxiliary.simulated_auxiliary
   :members:

.. automodule:: pykiso.lib.auxiliaries.simulated_auxiliary.simulation
   :members:

.. automodule:: pykiso.lib.auxiliaries.simulated_auxiliary.scenario
   :members:

.. automodule:: pykiso.lib.auxiliaries.simulated_auxiliary.response_templates
   :members:

"""

from . import simulated_auxiliary, simulation
from .simulated_auxiliary import SimulatedAuxiliary
from .simulation import Simulation



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
uds_request
***********

:module: uds_request

:synopsis: This module contains basics commands functions enums.

.. currentmodule:: uds_request
"""

from enum import Enum


class TupleEnum(tuple, Enum):
    """Base class for creating enumerated constants that are also
    subclasses of tuple.
    """

    pass


class UDSCommands:
    """Generic UDS commands defined by ISO 14229-1."""

    class TesterPresent(TupleEnum):
        """UDS requests to signal to the device that the client is still present"""

        TESTER_PRESENT = (0x3E, 0x00)
        TESTER_PRESENT_NO_RESPONSE = (0x3E, 0x80)

    class ECUReset(TupleEnum):
        """UDS requests to perform reset on components."""

        HARD_RESET = (0x11, 0x01)
        KEY_OFF_ON_RESET = (0x11, 0x02)
        SOFT_RESET = (0x11, 0x03)
        ENABLE_RAPID_POWER_SHUTDOWN = (0x11, 0x04)
        DISABLE_RAPID_POWER_SHUTDOWN = (0x11, 0x05)

    class Session(TupleEnum):
        """UDS requests to perform session change."""

        DEFAULT_SESSION = (0x10, 0x01)
        PROGRAMMING_SESSION = (0x10, 0x02)
        EXTENDED_SESSION = (0x10, 0x03)
        SAFTEFY_SYSTEM_SESSION = (0x10, 0x04)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Auxiliary interface
*******************

:module: aux_interface

:synopsis: Simply stored common methods for auxiliary's when ITF is used
    with Robot framework.

.. currentmodule:: aux_interface

"""
from robot.api import logger

from ...auxiliary import AuxiliaryInterface
from ...connector import CChannel, Flasher
from ...test_setup.config_registry import ConfigRegistry


class RobotAuxInterface:
    """Common interface for all Robot auxiliary."""

    def __init__(self, aux_type: AuxiliaryInterface):
        """Initialize attributes.

        :param aux_type: auxiliary's class
        """
        self.auxes = ConfigRegistry.get_auxes_by_type(aux_type)

    def _get_aux(self, alias: str) -> AuxiliaryInterface:
        """Return the corresponding auxiliary instance according to the
        given auxiliary's alias.

        :param alias: auxiliary's alias

        :return: auxiliary instance

        :raise: KeyError when auxiliary's alias is not part of auxes
            dictionary's keys.
        """
        if alias not in self.auxes.keys():
            logger.error(
                f"Auxiliary {alias} does not exist. Check your Yaml config",
                html=True,
            )
            raise KeyError(f"Auxiliary {alias} doesn't exist")
        return self.auxes[alias]

    def _get_aux_connectors(self, alias: str) -> CChannel:
        """Return the corresponding cchannel instance according to the
        given auxiliary's alias.

        :param alias: auxiliary's alias

        :return: associated CChannel instance
        """
        aux = self._get_aux(alias)
        return aux.channel

    def _get_aux_flasher(self, alias: str) -> Flasher:
        """Return the corresponding flasher instance according to the
        given auxiliary's alias.

        :param alias: auxiliary's alias

        :return: associated Flasher instance
        """
        aux = self._get_aux(alias)
        return aux.flash



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Library of instruments communicating via VISA
*********************************************

:module: lib_instruments

:synopsis: Dictionaries containing the appropriate SCPI commands for
    some instruments.

.. currentmodule:: lib_instruments

"""

REGISTERED_INSTRUMENTS = [
    "Elektro-Automatik",
    "Rohde&Schwarz",
]

SCPI_COMMANDS_DICT = {
    # General
    "IDENTIFICATION": {
        "query": {"DEFAULT": "*IDN?"},
        "write": {"DEFAULT": "NOT_AVAILABLE"},
    },
    "STATUS_BYTE": {
        "query": {"DEFAULT": "*STB?"},
        "write": {"DEFAULT": "NOT_AVAILABLE"},
    },
    "ALL_ERRORS": {
        "query": {
            "DEFAULT": "SYST:ERR:ALL?",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
        "write": {"DEFAULT": "NOT_AVAILABLE"},
    },
    "SELF_TEST": {
        "query": {"DEFAULT": "*TST?"},
        "write": {"DEFAULT": "*TST"},
    },
    # Remote control
    "REMOTE_CONTROL": {
        "query": {
            "DEFAULT": "SYST:LOCK:OWN?",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
        "write": {"DEFAULT": "SYST:LOCK", "Rohde&Schwarz": "NOT_AVAILABLE"},
    },
    # Output selection and enable
    "OUTPUT_CHANNEL": {
        "query": {
            "DEFAULT": "INST:NSEL?",
            "Elektro-Automatik": "NOT_AVAILABLE",
        },
        "write": {
            "DEFAULT": "INST:NSEL",
            "Elektro-Automatik": "NOT_AVAILABLE",
        },
    },
    "OUTPUT_ENABLE": {
        "query": {"DEFAULT": "OUTP?"},
        "write": {"DEFAULT": "OUTP"},
    },
    # Nominal values
    "NOMINAL_VOLTAGE": {
        "query": {
            "DEFAULT": "SYSTem:NOMinal:VOLTage?",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
        "write": {
            "DEFAULT": "NOT_AVAILABLE?",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
    },
    "NOMINAL_CURRENT": {
        "query": {
            "DEFAULT": "SYSTem:NOMinal:CURRent?",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
        "write": {
            "DEFAULT": "NOT_AVAILABLE?",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
    },
    "NOMINAL_POWER": {
        "query": {
            "DEFAULT": "SYSTem:NOMinal:POWer?",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
        "write": {
            "DEFAULT": "NOT_AVAILABLE",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
    },
    # Output measurement
    "MEASURE_VOLTAGE": {
        "query": {
            "DEFAULT": "MEASure:VOLTage?",
        },
        "write": {
            "DEFAULT": "NOT_AVAILABLE",
        },
    },
    "MEASURE_CURRENT": {
        "query": {
            "DEFAULT": "MEASure:CURRent?",
        },
        "write": {
            "DEFAULT": "NOT_AVAILABLE",
        },
    },
    "MEASURE_POWER": {
        "query": {
            "DEFAULT": "MEASure:POWer?",
        },
        "write": {"DEFAULT": "NOT_AVAILABLE"},
    },
    # Output desired values and limits
    "VOLTAGE": {
        "query": {"DEFAULT": "SOURce:VOLTage?"},
        "write": {"DEFAULT": "SOURce:VOLTage"},
    },
    "CURRENT": {
        "query": {"DEFAULT": "SOURce:CURRent?"},
        "write": {"DEFAULT": "SOURce:CURRent"},
    },
    "POWER": {
        "query": {
            "DEFAULT": "SOURce:POWer?",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
        "write": {"DEFAULT": "SOURce:POWer", "Rohde&Schwarz": "NOT_AVAILABLE"},
    },
    "VOLTAGE_LIMIT_LOW": {
        "query": {
            "DEFAULT": "SOURce:VOLTage:LIMit:LOW?",
            "Rohde&Schwarz": "VOLT:ALIM:LOW?",
        },
        "write": {
            "DEFAULT": "SOURce:VOLTage:LIMit:LOW",
            "Rohde&Schwarz": "VOLT:ALIM:LOW",
        },
    },
    "VOLTAGE_LIMIT_HIGH": {
        "query": {
            "DEFAULT": "SOURce:VOLTage:LIMit:HIGH?",
            "Rohde&Schwarz": "VOLTage:ALIM?",
        },
        "write": {
            "DEFAULT": "SOURce:VOLTage:LIMit:HIGH",
            "Rohde&Schwarz": "VOLTage:ALIM",
        },
    },
    "CURRENT_LIMIT_LOW": {
        "query": {
            "DEFAULT": "SOURce:CURRent:LIMit:LOW?",
            "Rohde&Schwarz": "CURRent:ALIM:LOW?",
        },
        "write": {
            "DEFAULT": "SOURce:CURRent:LIMit:LOW",
            "Rohde&Schwarz": "CURRent:ALIM:LOW",
        },
    },
    "CURRENT_LIMIT_HIGH": {
        "query": {
            "DEFAULT": "SOURce:CURRent:LIMit:HIGH?",
            "Rohde&Schwarz": "CURRent:ALIM?",
        },
        "write": {
            "DEFAULT": "SOURce:CURRent:LIMit:HIGH",
            "Rohde&Schwarz": "CURRent:ALIM",
        },
    },
    "POWER_LIMIT_HIGH": {
        "query": {
            "DEFAULT": "SOURce:POWer:LIMit:HIGH?",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
        "write": {
            "DEFAULT": "SOURce:POWer:LIMit:HIGH",
            "Rohde&Schwarz": "NOT_AVAILABLE",
        },
    },
}



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Communication Channel Via Udp
********************************

:module: cc_udp

:synopsis: Udp communication channel

.. currentmodule:: cc_udp


.. warning: Still under test
"""

import logging
import socket
from typing import Dict, Optional

from pykiso import connector

log = logging.getLogger(__name__)


class CCUdp(connector.CChannel):
    """UDP implementation of the coordination channel."""

    def __init__(self, dest_ip: str, dest_port: int, **kwargs):
        """Initialize attributes.

        :param dest_ip: destination ip address
        :param dest_port: destination port
        """
        # Initialize the super class
        super().__init__(**kwargs)
        # Initialize the socket
        self.dest_ip = dest_ip
        self.dest_port = dest_port
        self.udp_socket = None
        self.source_addr = None
        # Define the max length
        self.max_msg_size = 256
        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6

    def _cc_open(self) -> None:
        """Open the udp socket."""
        self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def _cc_close(self) -> None:
        """Close the udp socket."""
        self.udp_socket.close()

    def _cc_send(self, msg: bytes, **kwargs) -> None:
        """Send message using udp socket

        :param msg: message to send, should bytes.
        :param kwargs: not used
        """

        self.udp_socket.sendto(msg, (self.dest_ip, self.dest_port))

    def _cc_receive(self, timeout: float = 0.0000001) -> Dict[str, Optional[bytes]]:
        """Read message from socket.

        :param timeout: timeout applied on receive event

        :return: dictionary containing the received bytes if successful, otherwise None
        """
        self.udp_socket.settimeout(timeout or self.timeout)

        try:
            msg_received, self.source_addr = self.udp_socket.recvfrom(self.max_msg_size)

        # catch the errors linked to the socket timeout without blocking
        except BlockingIOError:
            log.internal_debug(f"encountered error while receiving message via {self}")
            return {"msg": None}
        except socket.timeout:
            log.internal_debug(f"encountered error while receiving message via {self}")
            return {"msg": None}
        except BaseException:
            log.exception(f"encountered error while receiving message via {self}")
            return {"msg": None}

        return {"msg": msg_received}



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Define some recurring typing definitions
"""

import pathlib
import sys
from types import TracebackType
from typing import Any, Dict, List, NewType, Optional, Tuple, Type, Union

if sys.version_info < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict

from . import message

PathType = Union[str, pathlib.Path]

MsgType = Union[message.Message, bytes, str]

ExcInfoType = Tuple[Type[BaseException], BaseException, TracebackType]

ProxyReturn = Union[
    Dict[str, Union[bytes, int]],
    Dict[str, Union[bytes, None]],
    Dict[str, Union[message.Message, None]],
    Dict[str, None],
]

AuxiliaryAlias = NewType("AuxiliaryAlias", str)
ConnectorAlias = NewType("ConnectorAlias", str)


class AuxiliaryConfig(TypedDict):
    connectors: Dict[str, ConnectorAlias]
    config: Dict[str, Optional[Dict[str, Any]]]
    type: str


class ConnectorConfig(TypedDict):
    config: Dict[str, Optional[Dict[str, Any]]]
    type: str


class SuiteConfig(TypedDict):
    suite_dir: str
    test_filter_pattern: str
    test_suite_id: int


class ConfigDict(TypedDict):
    auxiliaries: Dict[AuxiliaryAlias, AuxiliaryConfig]
    connectors: Dict[ConnectorAlias, ConnectorConfig]
    test_suite_list: List[SuiteConfig]


class OdxParamDict(TypedDict):
    parameter: str


class OdxRequestConfigDict(TypedDict):
    service: int
    data: OdxParamDict



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Lauterbach Flasher
******************

:module: flash_lauterbach

:synopsis: used to flash through lauterbach probe.

.. currentmodule:: flash_lauterbach

.. warning: Still under test
"""
import ctypes
import enum
import logging
import subprocess
import time

from pykiso import connector

log = logging.getLogger(__name__)


class ScriptState(enum.IntEnum):
    """Use to determine script command execution."""

    UNKNOWN = -1
    NOT_RUNNING = 0
    RUNNING = 1
    DIALOG_OPEN = 2


class MessageLineState(enum.IntEnum):
    """Use to determine Message reading command."""

    ERROR = 2
    ERROR_INFO = 16


class LauterbachFlasher(connector.Flasher):
    """Connector used to flash through one and only one
    Lauterbach probe using Trace32 as remote API.
    """

    def __init__(
        self,
        t32_exc_path: str = None,
        t32_config: str = None,
        t32_script_path: str = None,
        t32_api_path: str = None,
        port: str = None,
        node: str = "localhost",
        packlen: str = "1024",
        device: int = 1,
        **kwargs,
    ):
        """Initialize attributes with configuration data.

        :param t32_exc_path: full path of Trace32 app to execute
        :param t32_config: full path of Trace32 configuration file
        :param t32_script_path: full path to .cmm flash script to execute
        :param t32_api_path: full path of remote api
        :param port: port number used for UDP communication
        :param node: node name (default localhost)
        :param packlen: data pack length for UDP communication (default 1024)
        :param device: configure device number given by Trace32 (default 1)
        """
        self.t32_script_path = t32_script_path
        self.t32_api_path = t32_api_path
        self.t32_start_args = [t32_exc_path, "-c", t32_config]
        self.port = str(port)
        self.node = str(node)
        self.packlen = str(packlen)
        self.device = device
        self.t32_process = None
        self.t32_api = None
        self.loadup_wait_time = 5
        self.allowed_t32_errors = 10
        super().__init__(self.t32_script_path, **kwargs)

    def open(self) -> None:
        """Open UDP socket between ITF and Trace32 loaded app.

        The open command leads to the following sub-tasks execution:

            - Open a Trace32 app
            - Load remote API using ctypes
            - Configure UPD channel (Port/buffer size...)
            - Open UDP connection
            - Make a ping request
        """
        # load Trace32 remote api
        try:
            self.t32_api = ctypes.cdll.LoadLibrary(self.t32_api_path)
            log.internal_info("Trace32 remote API loaded")
        except Exception as e:
            log.exception("Unable to open Trace32")
            raise e
        # Looking for an active Trace32 process and close it
        if self.t32_api.T32_Init() == 0:
            # Quit properly the previous T32 instance
            cmd_status = self.t32_api.T32_Cmd("QUIT".encode("latin-1"))
            log.internal_debug(f"Previous process Trace32 closed with exit code:{cmd_status}")
            # Properly exit t32 api. Otherwise subsequential commands will fail
            self.t32_api.T32_Exit()

        # open a new Trace32 process
        try:
            self.t32_process = subprocess.Popen(self.t32_start_args)
            log.internal_debug(f"Trace32 process open with arguments {self.t32_start_args}")
        except Exception as e:
            log.exception("Unable to open Trace32")
            raise e

        # wait until Trace32 app is launched
        time.sleep(self.loadup_wait_time)

        # channel configuration
        self.t32_api.T32_Config(b"NODE=", self.node.encode("utf-8"))
        self.t32_api.T32_Config(b"PORT=", self.port.encode("utf-8"))
        self.t32_api.T32_Config(b"PACKLEN=", self.packlen.encode("utf-8"))

        # open UDP connection with Trace32
        if self.t32_api.T32_Init() == 0:
            self.t32_api.T32_Attach(self.device)
            log.internal_info(f"ITF connected on {self.node}:{self.port}")
        else:
            log.fatal(f"Unable to connect on port : {self.port}")
            raise Exception(f"Unable to connect on port : {self.port}")

    def flash(self) -> None:
        """Flash software using configured .cmm script.

        The Flash command leads to the following sub-tasks execution :
            - Send to Trace32 CD.DO internal command (execute script)
            - Wait until script is finished
            - Get script execution verdict

        :raise Exception: if Trace32 error occurred during flash.
        """

        # read actual flash script content for logging puprpose
        with open(file=self.t32_script_path, mode="r") as script:
            log.internal_debug(
                f"flash using script at {self.t32_script_path}, with the following commands : {script.read()}"
            )

        # run flash script
        cmd = f"CD.DO {self.t32_script_path}"
        log.internal_info(f"Call T32_Cmd {cmd}")
        request_state = self.t32_api.T32_Cmd(cmd.encode("utf-8"))

        if request_state < 0:
            log.fatal("TRACE32 Remote API communication error")
            raise Exception("TRACE32 Remote API communication error")
        else:
            # wait until script's end
            script_state = ctypes.c_int(ScriptState.UNKNOWN)
            request_state = 0
            error_count = 0
            while not script_state.value == ScriptState.NOT_RUNNING:
                request_state = self.t32_api.T32_GetPracticeState(ctypes.byref(script_state))
                log.internal_debug(
                    f"Called T32_GetPracticeState. request_state: {request_state} "
                    f"script_state: {script_state.value} request error: {error_count}"
                )
                if request_state != 0:
                    error_count += 1
                if error_count >= self.allowed_t32_errors:
                    raise RuntimeError(
                        f"Error during lauterbach script run. Script: {self.t32_script_path}\n"
                        f"Abort execution because lauterbach was unresponsive for {error_count} times"
                    )
                time.sleep(1)

        # reset target
        cmd = "SYStem.RESet"
        log.internal_info(f"Call T32_Cmd {cmd}")
        request_state = self.t32_api.T32_Cmd(cmd.encode("utf-8"))

        # get script execution verdict
        script_state = ctypes.c_uint16(-1)
        message = ctypes.create_string_buffer(256)
        request_state = self.t32_api.T32_GetMessage(ctypes.byref(message), ctypes.byref(script_state))

        msg = message.value.decode("utf-8")
        if (
            request_state == 0
            and not script_state.value == MessageLineState.ERROR
            and not script_state.value == MessageLineState.ERROR_INFO
        ):
            log.internal_info("flash procedure successful")
        else:
            log.fatal(f"An error occurred during flash,state : {script_state.value} -> {msg}")
            raise Exception(f"An error occurred during flash,state : {script_state.value} -> {msg}")

    def close(self) -> None:
        """Close UDP socket and shut down Trace32 App."""
        # Close Trace32 application
        self.t32_api.T32_Cmd("QUIT".encode("latin-1"))

        # close communication with Trace32
        exit_state = self.t32_api.T32_Exit()
        log.internal_info(f"Disconnect from Trace32 with state {exit_state}")

        # Process has to quit properly.
        # Otherwise lauterbach needs a power reset to rerun t32mppc.
        try:
            self.t32_process.wait(timeout=5)
        except Exception:
            log.internal_warning("Trace32 failed to exit properly")



##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Pykiso-specific hooks for pytest
********************************

:module: hooks

:synopsis: define hooks for test auxiliary fixture customization.

"""

from __future__ import annotations

import pytest

from pykiso.interfaces.dt_auxiliary import DTAuxiliaryInterface


@pytest.hookspec(firstresult=True)
def pytest_auxiliary_load(aux: str) -> DTAuxiliaryInterface | None:
    """:meta private:"""


@pytest.hookspec(firstresult=True)
def pytest_auxiliary_start(aux: DTAuxiliaryInterface) -> bool | None:
    """Called to start an auxiliary.

    Stops at first non-None result.
    The return value is not used, but only stops further processing.

    :param aux: auxiliary instance that should be started.
    """


@pytest.hookspec(firstresult=True)
def pytest_auxiliary_stop(aux: DTAuxiliaryInterface) -> bool | None:
    """Called to stop an auxiliary.

    Stops at first non-None result.
    The return value is not used, but only stops further processing.

    :param aux: auxiliary instance that should be stopped.
    """



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Execution of tests
******************

:module: test_execution

:synopsis: Execute a test environment based on the supplied configuration.

.. currentmodule:: test_execute

.. note::
    1. Glob a list of test-suite folders
    2. Generate a list of test-suites with a list of test-cases
    3. Loop per suite
    4. Gather result
"""
from __future__ import annotations

import functools
import os
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Literal, Optional, Union
from unittest import util
from unittest.loader import VALID_MODULE_NAME

import pykiso.test_result.multi_result as multi_result

if TYPE_CHECKING:
    from ..lib.connectors.cc_pcan_can import CCPCanCan
    from ..lib.connectors.cc_socket_can.cc_socket_can import CCSocketCan
    from ..types import ConfigDict, SuiteConfig
    from .test_case import BasicTest

import enum
import logging
import re
import time
import unittest
from collections import OrderedDict, namedtuple
from datetime import datetime
from fnmatch import fnmatch
from pathlib import Path

import xmlrunner

import pykiso

from ..exceptions import AuxiliaryCreationError, TestCollectionError
from ..logging_initializer import get_logging_options
from ..test_result.assert_step_report import StepReportData, assert_decorator, generate_step_report
from ..test_result.text_result import BannerTestResult, ResultStream
from ..test_result.xml_result import XmlTestResult
from . import test_suite

log = logging.getLogger(__name__)

TestFilterPattern = namedtuple("TestFilterPattern", "test_file, test_class, test_case")


@enum.unique
class ExitCode(enum.IntEnum):
    """List of possible exit codes"""

    ALL_TESTS_SUCCEEDED = 0
    ONE_OR_MORE_TESTS_FAILED = 1
    ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION = 2
    ONE_OR_MORE_TESTS_FAILED_AND_RAISED_UNEXPECTED_EXCEPTION = 3
    AUXILIARY_CREATION_FAILED = 4
    BAD_CLI_USAGE = 5
    UNRESOLVED_THREADS = 6


def create_test_suite(
    test_suite_dict: SuiteConfig,
) -> test_suite.BasicTestSuite:
    """create a test suite based on the config dict

    :param test_suite_dict: dict created from config with keys 'suite_dir',
        'test_filter_pattern', 'test_suite_id'
    """
    return test_suite.BasicTestSuite(
        test_suite_dict["suite_dir"],
        test_suite_dict["test_filter_pattern"],
        test_suite_dict["test_suite_id"],
    )


def apply_tag_filter(all_tests_to_run: unittest.TestSuite, usr_tags: Dict[str, List[str]]) -> None:
    """Filter the test cases based on user tags provided via CLI.

    :param all_tests_to_run: a dict containing all testsuites and testcases
    :param usr_tags: encapsulate user's variant choices
    """

    def format_tag_names(tags: Dict[str, List[str]]) -> Dict[str, List[str]]:
        """Remove any comma or underscore from the provided dict's keys."""
        for char in ["_", "-"]:
            tags = {name.replace(char, ""): value for name, value in tags.items()}
        return tags

    def should_skip(test_case: BasicTest) -> bool:
        """Check if test shall be skipped by evaluating the test case tag
        attribute.

        :param test_case: test_case to check
        :return: True if test shall be skipped else False
        """
        if test_case.tag is None:
            return False

        # Skip only the test cases that have a matching tag name but no matching tag value
        for cli_tag_id, cli_tag_value in usr_tags.items():
            # skip any test case that doesn't define a CLI-provided tag name
            if cli_tag_id not in test_case.tag.keys():
                test_case._skip_msg = f"provided tag {cli_tag_id!r} not present in test tags"  # noqa: E713
                return True
            # skip any test case that which tag value don't match the provided tag's value
            cli_tag_values = cli_tag_value if isinstance(cli_tag_value, list) else [cli_tag_value]
            if not any(cli_val in test_case.tag[cli_tag_id] for cli_val in cli_tag_values):
                test_case._skip_msg = f"non-matching value for tag {cli_tag_id!r}"
                return True

        return False

    def set_skipped(test_case: BasicTest) -> BasicTest:
        """Set testcase to skipped.

        :param test_case: testcase to be skipped
        :return: the skipped testcase as a new instance of the provided
            TestCase subclass.
        """
        if should_skip(test_case):
            skipped_test_cls = unittest.skip(test_case._skip_msg)(test_case.__class__)
            return skipped_test_cls()
        return test_case

    # collect and reformat all CLI and test case tag names
    usr_tags = format_tag_names(usr_tags)

    all_test_tags = []
    for tc in test_suite.flatten(all_tests_to_run):
        if getattr(tc, "tag", None) is None:
            continue
        tc.tag = format_tag_names(tc.tag)
        all_test_tags.extend(tc.tag.keys())

    # skip the tests according to the provided CLI tags and the defined test tags
    list(map(set_skipped, test_suite.flatten(all_tests_to_run)))

    # verify that each provided tag name is defined in at least one test case
    all_test_tags = set(all_test_tags)
    for tag_name in usr_tags:
        if tag_name not in all_test_tags:
            raise NameError(
                f"Provided tag {tag_name!r} is not defined in any testcase.",
                tag_name,
            )


def apply_test_case_filter(
    all_tests_to_run: unittest.TestSuite,
    test_class_pattern: str,
    test_case_pattern: str,
) -> unittest.TestSuite:
    """Apply a filter to run only test cases which matches given expression

    :param all_tests_to_run: a dict containing all testsuites and testcases
    :param test_class_pattern: pattern to select test class as unix filename pattern
    :param test_case_pattern: pattern to select test case as unix filename pattern
    :return: new test suite with filtered test cases
    """

    def is_active_test(test_case: BasicTest) -> bool:
        """Check if testcase shall be active by given selection patterns

        :param test_case: unittest test
        :return: True if test matches patterns else False
        """
        test_class_name = re.sub(r"-\d+-\d+", "", test_case.__class__.__name__)
        is_classname_match = bool(fnmatch(test_class_name, test_class_pattern))

        if is_classname_match and test_case_pattern is None:
            return is_classname_match
        elif is_classname_match and test_case_pattern:
            return bool(fnmatch(test_case._testMethodName, test_case_pattern))
        else:
            return False

    base_suite = test_suite.flatten(all_tests_to_run)
    filtered_suite = filter(is_active_test, base_suite)
    return unittest.TestSuite(filtered_suite)


def failure_and_error_handling(result: unittest.TestResult) -> int:
    """provide necessary information to Jenkins if an error occur during tests execution

    :param result: encapsulate all test results from the current run

    :return: an ExitCode object
    """
    failed, errored = result.failures, result.errors
    if failed and errored:
        exit_code = ExitCode.ONE_OR_MORE_TESTS_FAILED_AND_RAISED_UNEXPECTED_EXCEPTION
    elif failed:
        exit_code = ExitCode.ONE_OR_MORE_TESTS_FAILED
    elif errored:
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    else:
        exit_code = ExitCode.ALL_TESTS_SUCCEEDED
    return exit_code


def enable_step_report(all_tests_to_run: unittest.suite.TestSuite, step_report: Path) -> None:
    """Decorate all assert method from Test-Case.

    This will allow to save the assert inputs in
    order to generate the HTML step report.

    :param all_tests_to_run: a dict containing all testsuites and testcases
    """

    # Step report header fed during test
    base_suite = test_suite.flatten(all_tests_to_run)
    for tc in base_suite:
        tc.step_report = StepReportData()
        if step_report is not None:
            # for any test, show ITF version
            tc.step_report.header = OrderedDict({"ITF version": pykiso.__version__})
            # Decorate All assert method
            assert_method_list = [method for method in dir(tc) if method.startswith("assert")]
            for method_name in assert_method_list:
                # Get method from name
                method = getattr(tc, method_name)
                # Add decorator to the existing method
                setattr(tc, method_name, assert_decorator(method))


def parse_test_selection_pattern(pattern: str) -> TestFilterPattern:
    """Parse test selection pattern from cli.

    For example: ``test_file.py::TestCaseClass::test_method``

    :param pattern: test selection pattern
    :return: pattern for file, class name and test case name
    """
    if not pattern:
        return TestFilterPattern(None, None, None)
    parsed_patterns = []
    patterns = pattern.split("::")
    for pattern in patterns:
        if pattern == "":
            parsed_patterns.append(None)
        else:
            parsed_patterns.append(pattern)
    for _ in range(3 - (len(parsed_patterns))):
        parsed_patterns.append(None)

    return TestFilterPattern(*parsed_patterns)


def find_folders_between_paths(start_path: Path, end_path: Path) -> List[Path]:
    """Find and return a list of folders between two specified paths.

    :param start_path: the starting path
    :param end_path: the ending path
    :return: a list of folders between `start_path` (exclusive) and `end_path` (inclusive). If `end_path` is not
        a subpath of `start_path`, an empty list is returned.
    """
    start_path = start_path
    end_path = end_path
    folders_between = []

    while start_path != end_path:
        folders_between.append(end_path.parent)
        end_path = end_path.parent
        if len(end_path.parents) == 1:
            return []
    return folders_between


def _is_valid_module(start_dir: str, pattern: str) -> bool:
    """Check if the test files found in the specified directory and its subdirectories
    conform to the requirements of a valid module.

    :param start_dir: the directory to search
    :param pattern: pattern that matches the file names
    """
    start_dir = Path(start_dir)
    test_files_found = list(start_dir.glob(f"**/{pattern}"))

    # remove folders
    test_files_found = [test_file for test_file in test_files_found if test_file.is_file()]

    # If found test files are in given test suite directory don't check for __init__.py
    if any(test_file.parent == start_dir for test_file in test_files_found):
        return True

    # No file matches the test filter pattern
    if not test_files_found:
        log.critical(f"Test filter {pattern=} doesn't match any files in folder {start_dir}")
        return False

    # Check that all sub folders of a nested testsuite have an __init__.py file
    testsuite_folders = {filepath.parent for filepath in test_files_found}
    all_folders_to_check = []
    for folder in testsuite_folders:
        all_folders_to_check.extend(find_folders_between_paths(start_dir, folder))
    all_folders_to_check.extend(testsuite_folders)
    all_folders_to_check = list(set(all_folders_to_check))

    if start_dir in all_folders_to_check:
        all_folders_to_check.remove(start_dir)

    for folder in all_folders_to_check:
        if not any(file_path.name == "__init__.py" for file_path in folder.iterdir()):
            log.critical(f'Could not find file "__init__.py" in folder {folder}')
            return False

    # check that the test file has a valid module name
    return all(VALID_MODULE_NAME.match(file.name) for file in test_files_found)


def filter_test_modules_by_suite(
    test_modules: List[SuiteConfig],
) -> List[SuiteConfig]:
    """Filter a list of test modules by their suite directory to avoid running duplicates.

    :param test_modules: List of test modules to filter.
    :return: Filtered list of test modules with unique suite directories
    """
    seen_suite_dirs = {}
    filtered_data = []

    for module in test_modules:
        suite_dir = module["suite_dir"]

        # Check if the suite directory is seen for the first time
        if suite_dir not in seen_suite_dirs:
            seen_suite_dirs[suite_dir] = module
            filtered_data.append(module)

    return filtered_data


def collect_test_suites(
    config_test_suite_list: List[SuiteConfig],
    test_filter_pattern: Optional[str] = None,
) -> List[test_suite.BasicTestSuite]:
    """Collect and load all test suites defined in the test configuration.

    :param config_test_suite_list: list of dictionaries from the configuration
        file corresponding each to one test suite.
    :param test_filter_pattern: optional filter pattern to overwrite
        the one defined in the test suite configuration.

    :raises pykiso.TestCollectionError: if any test case inside one of
        the configured test suites failed to be loaded.

    :return: a list of all loaded test suites.
    """
    list_of_test_suites = []
    valid_test_modules = []

    for test_suite_configuration in config_test_suite_list:
        if test_filter_pattern is not None:
            test_suite_configuration["test_filter_pattern"] = test_filter_pattern

        if _is_valid_module(
            start_dir=test_suite_configuration["suite_dir"],
            pattern=test_suite_configuration["test_filter_pattern"],
        ):
            valid_test_modules.append(test_suite_configuration)

    if test_filter_pattern:
        valid_test_modules = filter_test_modules_by_suite(valid_test_modules)

    if not valid_test_modules:
        raise TestCollectionError([test_suite_config["suite_dir"] for test_suite_config in config_test_suite_list])

    for test_suite_configuration in valid_test_modules:
        try:
            current_test_suite = create_test_suite(test_suite_configuration)
            list_of_test_suites.append(current_test_suite)
        except BaseException as e:
            raise TestCollectionError(test_suite_configuration["suite_dir"]) from e
    return list_of_test_suites


def abort(reason: str = None) -> None:
    """Quit ITF test and log an error if a reason is indicated and
    if any errors occurred it logs them.

    :param reason: reason to abort, defaults to None
    """
    if reason:
        log.critical(reason)
    log.critical("Non recoverable error occurred during test execution, aborting test session.")
    os.kill(os.getpid(), ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION)


def execute(
    config: ConfigDict,
    report_type: str = "text",
    report_name: str = "",
    user_tags: Optional[Dict[str, List[str]]] = None,
    step_report: Optional[Path] = None,
    pattern_inject: Optional[str] = None,
    failfast: bool = False,
    junit_path: str = "reports",
    log_file_strategy: Literal["testRun", "testCase"] | None = None,
) -> int:
    """Create test environment based on test configuration.

    :param config: dict from converted YAML config file
    :param report_type: str to set the type of report wanted, i.e. test
        or junit
    :param report_name: name of the junit report
    :param user_tags: test case tags to execute
    :param step_report: file path for the step report or None
    :param pattern_inject: optional pattern that will override
        test_filter_pattern for all suites. Used in test development to
        run specific tests.
    :param failfast: stop the test run on the first error or failure.
    :param junit_path: path (file or dir) to junit report
    :param log_file_strategy: strategy for the log file creation

    :return: exit code corresponding to the result of the test execution
        (tests failed, unexpected exception, ...)
    """
    try:
        test_file_pattern = parse_test_selection_pattern(pattern_inject)

        test_suites = collect_test_suites(config["test_suite_list"], test_file_pattern.test_file)
        test_suites = handle_can_trace_strategy(config, test_suites)
        # Group all the collected test suites in one global test suite
        all_tests_to_run = unittest.TestSuite(test_suites)

        # filter test cases based on variant and branch-level options
        if user_tags:
            apply_tag_filter(all_tests_to_run, user_tags)

        # Enable step report
        enable_step_report(all_tests_to_run, step_report)

        if test_file_pattern.test_class:
            all_tests_to_run = apply_test_case_filter(
                all_tests_to_run,
                test_file_pattern.test_class,
                test_file_pattern.test_case,
            )

        log_file_path = get_logging_options().log_path
        current_time = datetime.today()
        # TestRunner selection: generate or not a junit report. Start the tests and publish the results
        if report_type == "junit":
            report_path = junit_path
            full_report_path = Path.cwd() / report_path
            if full_report_path.suffix == ".xml":
                junit_report_path = str(full_report_path)
                full_report_path.parent.mkdir(exist_ok=True)
            else:
                junit_report_path = str(
                    full_report_path / Path(current_time.strftime(f"%Y-%m-%d_%H-%M-%S-{report_name}.xml"))
                )
                full_report_path.mkdir(exist_ok=True)
            with open(junit_report_path, "wb") as junit_output, ResultStream(log_file_path) as stream:
                multi_result.test_runner_instance = xmlrunner.XMLTestRunner(
                    output=junit_output,
                    resultclass=multi_result.MultiTestResult(
                        XmlTestResult, BannerTestResult, log_file_strategy=log_file_strategy
                    ),
                    failfast=failfast,
                    verbosity=0,
                    stream=stream,
                )
                result = multi_result.test_runner_instance.run(all_tests_to_run)
                # Generate the html step report
                if step_report is not None:
                    generate_step_report(result, step_report)

        else:
            with ResultStream(log_file_path) as stream:
                multi_result.test_runner_instance = unittest.TextTestRunner(
                    stream=stream,
                    resultclass=multi_result.MultiTestResult(BannerTestResult, log_file_strategy=log_file_strategy),
                    failfast=failfast,
                    verbosity=0,
                )
                result = multi_result.test_runner_instance.run(all_tests_to_run)
                # Generate the html step report
                if step_report is not None:
                    generate_step_report(result, step_report)

        exit_code = failure_and_error_handling(result)
    except NameError:
        log.exception("Error occurred during tag evaluation.")
        exit_code = ExitCode.BAD_CLI_USAGE
    except TestCollectionError:
        log.exception("Error occurred during test collection.")
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    except AuxiliaryCreationError:
        log.exception("Error occurred during auxiliary creation.")
        exit_code = ExitCode.AUXILIARY_CREATION_FAILED
    except KeyboardInterrupt:
        log.exception("Keyboard Interrupt detected")
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    except Exception:
        log.exception(f'Issue detected in the test-suite: {config["test_suite_list"]}!')
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    return int(exit_code)


def handle_can_trace_strategy(
    config: dict[str, Any], test_suites: list[unittest.TestSuite]
) -> list[unittest.TestSuite]:
    """Setup the test to get a trace file for every test or testCase if the user requested it
        else just return the test suite passed in input

    :param config: dict from converted YAML config file
    :return: a list of all loaded test suites.

    :return: a list of all loaded test suites with setUp and tearDown modified if needed.
    """
    if not _check_trace_file_strategy(config):
        return test_suites
    can_channel = _get_connector_instance_with_trace_file_strategy(config)
    if can_channel is None:
        return test_suites

    # Decorate function to call the start and stop pcan trace function
    list_class: list[type[unittest.TestCase]] = []
    for suite in test_suites:
        for test in suite._tests:
            # If the user want to have one trace file for a testCase we decorate the setUpClass and tearDownClass
            if can_channel.strategy_trc_file == "testCase":
                trc_file_name = util.strclass(test.__class__).replace(".", "_").replace("-", "_") + ".trc"
                test_class = test.__class__
                if test_class in list_class:
                    continue
                list_class.append(test_class)
                test_class.setUpClass = start_can_trace_decorator(test_class.setUpClass, can_channel, trc_file_name)
                test_class.tearDownClass = stop_can_trace_decorator(test_class.tearDownClass, can_channel)

            # If the user want to have one trace file for each test run we decorate the setUp and tearDown of the class
            elif can_channel.strategy_trc_file == "testRun":
                trc_file_name = (
                    util.strclass(test.__class__).replace(".", "_").replace("-", "_")
                    + "_"
                    + test._testMethodName
                    + ".trc"
                )
                test.setUp = start_can_trace_decorator(test.setUp, can_channel, trc_file_name)
                test.tearDown = stop_can_trace_decorator(test.tearDown, can_channel)
    return test_suites


def _get_connector_instance_with_trace_file_strategy(config: dict[str, Any]) -> Optional[Union[CCPCanCan, CCSocketCan]]:
    """Get pcan and socket can channels from auxiliaries created from the Yaml

    :param config: dict from converted YAML config file
    :return: the channel instance if one was created else False
    """
    from ..lib.connectors.cc_pcan_can import CCPCanCan
    from ..lib.connectors.cc_proxy import CCProxy
    from ..lib.connectors.cc_socket_can import CCSocketCan

    # Retrieve the channel from the auxiliaries defined
    for aux in config["auxiliaries"].keys():
        instance_channel = getattr(getattr(pykiso.auxiliaries, aux, None), "channel", None)
        if isinstance(instance_channel, CCProxy):
            instance_channel = instance_channel._proxy.channel
        if isinstance(instance_channel, (CCPCanCan, CCSocketCan)) and instance_channel.strategy_trc_file in [
            "testRun",
            "testCase",
        ]:
            return instance_channel
    return None


def _check_trace_file_strategy(config: dict[str, Any]) -> bool:
    """Check if at least one connector has the trace file strategy flag active

    :param config: dict from converted YAML config file

    :return: True if a connector has a trace file strategy active else False
    """
    strategy_trc_file = None
    # Check if a pcan with the right parameter is defined in the config
    for connector in config["connectors"].values():
        connector_config = connector.get("config", {})
        if connector_config is not None and "strategy_trc_file" in connector.get("config", {}).keys():
            strategy_trc_file = connector["config"]["strategy_trc_file"]
            if strategy_trc_file not in ["testRun", "testCase"]:
                raise ValueError(
                    f"{strategy_trc_file} is not a valid value for strategy_trc_file, ",
                    "valid values are ['testRun','testCase']",
                )
    return strategy_trc_file is not None


def start_can_trace_decorator(func: Callable, channel: Union[CCPCanCan, CCSocketCan], trace_file_name: str):
    """Decorator that will call start pcan trace before calling the function

    :param func: function to execute
    :param cc_pcan: channel used to get the pcan trace
    :param trace_file_name: name of the trace file to create
    """

    @functools.wraps(func)
    def decorator(*args, **kwargs):
        from ..lib.connectors.cc_pcan_can import CCPCanCan

        # Add datetime in trace file name to not overwrite trace file when rerunning test
        if not channel.opened:
            channel.open()
        trace_path = channel.trace_path / trace_file_name.replace(
            ".trc", f"_{datetime.today().strftime('%Y%d%m%H%M%S')}.trc"
        )
        if isinstance(channel, CCPCanCan):
            channel.start_pcan_trace(trace_path=trace_path)
        else:
            channel.start_can_trace(trace_path=trace_path)
        return func(*args, **kwargs)

    return decorator


def stop_can_trace_decorator(func: Callable, channel: Union[CCPCanCan, CCSocketCan]):
    """Decorator that will call stop pcan trace after calling the function

    :param func: function to execute
    :param channel: channel used to get the pcan trace
    """

    @functools.wraps(func)
    def decorator(*args, **kwargs):
        from ..lib.connectors.cc_pcan_can import CCPCanCan

        result = func(*args, **kwargs)
        if isinstance(channel, CCPCanCan):
            channel.stop_pcan_trace()
        else:
            channel.stop_can_trace()
        return result

    return decorator



##########################################################################
# Copyright (c) 2023-2023 Accenture GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Communication Channel Via GRPC
******************************

:module: cc_grpc_client

:synopsis: Grpc communication channel

.. currentmodule:: cc_grpc_client

list of todo items:
* Extend by supporting a path instead of files so that many proto files can be used
* Extend with security capabilities

"""

import importlib.util
import inspect
import logging
from pathlib import Path
from typing import Dict, Optional

import grpc

from pykiso import connector

log = logging.getLogger(__name__)


class CCGrpcClient(connector.CChannel):
    """Grpc client implementation"""

    def __init__(
        self,
        dest_ip: str,
        dest_port: int,
        generated_protobuf_data_types_file: str,
        generated_protobuf_grpc_file: str,
        default_service_name: str,
        default_rpc_name: str,
        default_message_name: str,
        default_message_fields: dict,
        **kwargs,
    ):
        """Instantiate a grpc client

        :param dest_ip (str): grpc server ip address
        :param dest_port (int): grpc server port
        :param generated_protobuf_data_types_file (str): path to the generated protobuf data types file
        :param generated_protobuf_grpc_file (str): path to the generated protobuf grpc file
        :param default_service_name (str): default service name to use
        :param default_rpc_name (str): default rpc name to use
        :param default_message_name (str): default message name to use
        :param default_message_fields (dict): default parameters to use in dictionary format
        """
        # Initialize the super class
        super().__init__(**kwargs)
        # Initialize the grpc client
        self.dest_ip = dest_ip
        self.dest_port = dest_port
        # Import the generated datamodel protobuf file
        spec = importlib.util.spec_from_file_location(
            Path(generated_protobuf_data_types_file).stem,
            Path(generated_protobuf_data_types_file),
        )
        self.grpc_datamodel = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.grpc_datamodel)
        # Import the generated grpc protobuf file
        # Import the module
        spec = importlib.util.spec_from_file_location(
            Path(generated_protobuf_grpc_file).stem,
            Path(generated_protobuf_grpc_file),
        )
        self.grpc_stub_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.grpc_stub_module)
        # Extract the stubs
        module_attributes = inspect.getmembers(self.grpc_stub_module, inspect.isclass)
        self.grpc_stub = []
        for attribute in module_attributes:
            if attribute[0].endswith("Stub"):
                self.grpc_stub.append(attribute[1])
        # Prepare the empty stub list. It will be used for binding the stubs to the channel
        self.stubs = {}

        # Set the default rpc and message
        self.default_service_name = default_service_name
        self.default_rpc_name = default_rpc_name
        self.default_message_name = default_message_name
        self.default_message_fields = default_message_fields
        self.list_of_received_messages = []

    def _cc_open(self) -> None:
        """Open the grpc connection to server."""
        self.channel = grpc.insecure_channel(self.dest_ip + ":" + str(self.dest_port))
        # Bind stubs to new channel
        self.stubs = dict([(stub.__name__[:-4], stub(self.channel)) for stub in self.grpc_stub])
        # Validate at least that most of the default parameters are valid
        try:
            if self.default_service_name not in self.stubs:
                raise ValueError(f"{self.default_service_name=} was not found!")
            if not hasattr(self.stubs[self.default_service_name], self.default_rpc_name):
                raise ValueError(f"{self.default_rpc_name=} was not found!")
            if not hasattr(self.grpc_datamodel, self.default_message_name):
                raise ValueError(f"{self.default_message_name=} was not found!")
            message_class = getattr(self.grpc_datamodel, self.default_message_name)
            message_class(**self.default_message_fields)
        except TypeError as e:
            self._cc_close()
            raise ValueError(f"{self.default_message_fields=} was not valid!") from e
        except Exception as e:
            self._cc_close()
            raise e

    def _cc_close(self) -> None:
        """Close the grpc connection to server."""
        self.stubs.clear()
        self.channel.close()

    def _cc_send(self, msg: bytes, **kwargs) -> None:
        """Send message to grpc server.

        :param msg: message to send, should bytes.
        :param kwargs: used to extract the service_name, rpc_name, message_name and message_fields

        In addition to the message, the following keyword arguments are supported:

            - rpc_name: name of the rpc to trigger
            - message_name: name of the message to use
            - message_fields: attributes to use in dictionary format

        """
        # Extract needed parameters or the default once
        service_name = kwargs.get("service_name", self.default_service_name)
        rpc_name = kwargs.get("rpc_name", self.default_rpc_name)
        message_name = kwargs.get("message_name", self.default_message_name)
        # For the message attributes, we have 3 possibilities:
        # * msg is populated OR
        # * message_fields is populated OR
        # * the default message attributes are used
        message_fields = msg if msg else kwargs.get("message_fields", self.default_message_fields)

        # Now, the validity of the parameters need to be checked
        # First we check that the service_name exist
        if service_name not in self.stubs:
            log.warning(f"{service_name=} was not found! Fallback to {self.default_service_name=}")
            service_name = self.default_service_name
        # Then we check that the rpc_name exist
        if not hasattr(self.stubs[service_name], rpc_name):
            log.warning(f"{rpc_name=} was not found! Fallback to {self.default_rpc_name=}")
            rpc_name = self.default_rpc_name
        # Then we check that the message_name exist
        if not hasattr(self.grpc_datamodel, message_name):
            log.warning(f"{message_name=} was not found! Fallback to {self.default_message_name=}")
            message_name = self.default_message_name
        # The message constitution will be checked later on, when the message will be sent

        # Trigger the rpc
        rpc_method = getattr(self.stubs[service_name], rpc_name)
        message_class = getattr(self.grpc_datamodel, message_name)
        result = self._rpc_trigger(rpc_method, message_class, message_fields)

        # Store the result
        self.list_of_received_messages.append(result)

    def _rpc_trigger(self, rpc_method: str, message_class: str, message_fields: dict = None) -> None:
        """Trigger the rpc.

        :param rpc_method: rpc method to trigger
        :param message_class: message class to use
        :param message_fields: attributes to use in dictionary format

        :return: result of the rpc
        """
        try:
            if message_fields:
                return rpc_method(message_class(**message_fields))
            else:
                return rpc_method(message_class())
        except Exception as e:
            log.exception(f"Error while triggering rpc: {e}")
            return None

    def _cc_receive(self, timeout: float = 1e-6) -> Dict[str, Optional[bytes]]:
        """Read message from socket.

        :param timeout: timeout applied on receive event

        :return: dictionary containing the received bytes if successful, otherwise None
        """

        try:
            return {"msg": self.list_of_received_messages.pop(0)}
        except IndexError:
            log.debug(f"no message received via {self}")
            return {"msg": None}



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Robotframework Libraries
************************

:module: robotframework

.. currentmodule:: robotframework

"""



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Testapp binding
***************

:module: dut_auxiliary

:synopsis: implementation of existing DUTAuxiliary for Robot
    framework usage.

.. currentmodule:: dut_auxiliary

"""

from typing import Any, Callable, List, Tuple

from robot.api import logger
from robot.api.deco import keyword, library

from pykiso.test_coordinator.test_case import RemoteTest
from pykiso.test_coordinator.test_suite import RemoteTestSuiteSetup, RemoteTestSuiteTeardown

from ..auxiliaries.dut_auxiliary import COMMAND_TYPE
from ..auxiliaries.dut_auxiliary import DUTAuxiliary as DutAux
from .aux_interface import RobotAuxInterface


@library(version="0.0.1")
class DUTAuxiliary(RobotAuxInterface):
    """Robot library to control the TestApp on the DUT"""

    ROBOT_LIBRARY_SCOPE = "SUITE"

    def __init__(self) -> None:
        """Initialize attributes."""
        super().__init__(aux_type=DutAux)
        self.test_entity_mapping = {
            COMMAND_TYPE.TEST_SUITE_SETUP: (
                RemoteTestSuiteSetup,
                "test_suite_setUp",
            ),
            COMMAND_TYPE.TEST_SUITE_TEARDOWN: (
                RemoteTestSuiteTeardown,
                "test_suite_tearDown",
            ),
            COMMAND_TYPE.TEST_CASE_SETUP: (RemoteTest, "setUp"),
            COMMAND_TYPE.TEST_CASE_RUN: (RemoteTest, "test_run"),
            COMMAND_TYPE.TEST_CASE_TEARDOWN: (RemoteTest, "tearDown"),
        }

    def _get_test_class_info(self, cmd_alias: str) -> Tuple[Any, Callable]:
        """Retrieve the corresponding couple test class/ test fixture
        based on the given command name.

        :param cmd_alias: command name or alias (TEST_CASE_RUN,
            TEST_CASE_SETUP...)

        :return: test class object to insatnciate and fixture name to
            call

        :raises TypeError: if the given command type doesn't exist
        """
        try:
            command = COMMAND_TYPE[cmd_alias]
        except KeyError:
            raise TypeError(f"unknown command type {cmd_alias}")

        return self.test_entity_mapping.get(command)

    @keyword(name="Test App")
    def test_app_run(
        self,
        command_type: str,
        test_suite_id: int,
        test_case_id: int,
        aux_list: List[str],
    ) -> None:
        """Execute the corresponding test fixture using Test App
        communication protocol.

        :param command_type: message command sub-type
        :param test_suite_id: select test suite id on dut
        :param test_case_id: select test case id on dut
        :param aux_list: List of selected auxiliary

        :raises TypeError: if the given command type doesn't exist
        :raises Assertion: if an aknowledgment is not received or the
            report status is failed.
        """

        auxiliaries = [self._get_aux(aux) for aux in aux_list]

        test_cls, fixture_alias = self._get_test_class_info(command_type)

        test_instance = test_cls(
            test_suite_id,
            test_case_id,
            auxiliaries,
            setup_timeout=None,
            run_timeout=None,
            teardown_timeout=None,
            test_ids=None,
            tag=None,
        )

        fixture_func = getattr(test_instance, fixture_alias)

        fixture_func()
